"""
Image depth
===========

.. tags:: visualization-basic
"""

import numpy as np

import napari

im_data = np.zeros((50, 50, 50))
im_data[30:40, 25:35, 25:35] = 1
viewer = napari.Viewer()
layer = viewer.add_image(im_data, colormap='magenta', rendering='iso')
viewer.add_image(im_data, colormap='green', rendering='iso', translate=(30, 0, 0))

points_data = [
    [50, 30, 30],
    [25, 30, 30],
    [75, 30, 30]
]
viewer.add_points(points_data, size=4)

viewer.dims.ndisplay = 3

if __name__ == '__main__':
    napari.run()



"""
Script to check for string in the codebase not using `trans`.

TODO:
  * Find all logger calls and add to skips
  * Find nested funcs inside if/else


Run manually with

    $ pytest -Wignore tools/ --tb=short

To interactively be prompted whether new strings should be ignored or need translations.


You can pass a command to also have the option to open your editor.
Example here to stop in Vim at the right file and linenumber.


    $ python tools/test_strings.py "vim {filename} +{linenumber}"
"""

import ast
import os
import subprocess
import sys
import termios
import tokenize
import tty
from contextlib import suppress
from pathlib import Path
from types import ModuleType
from typing import Optional

import pytest
from strings_list import (
    SKIP_FILES,
    SKIP_FOLDERS,
    SKIP_WORDS,
    SKIP_WORDS_GLOBAL,
)

REPO_ROOT = Path(__file__).resolve()
NAPARI_MODULE = (REPO_ROOT / 'napari').relative_to(REPO_ROOT)

# Types
StringIssuesDict = dict[str, list[tuple[int, str]]]
OutdatedStringsDict = dict[str, list[str]]
TranslationErrorsDict = dict[str, list[tuple[str, str]]]


class FindTransStrings(ast.NodeVisitor):
    """This node visitor finds translated strings."""

    def __init__(self) -> None:
        super().__init__()

        self._found = set()
        self._trans_errors = []

    def _check_vars(self, method_name, args, kwargs):
        """Find interpolation variables inside a translation string.

        This helps find any variables that need to be interpolated inside
        a string so we can check against the `kwargs` for both singular
        and plural strings (if present) inside `args`.

        Parameters
        ----------
        method_name : str
            Translation method used. Options include "_", "_n", "_p" and
            "_np".
        args : list
            List of arguments passed to translation method.
        kwargs : kwargs
            List of keyword arguments passed to translation method.
        """
        singular_kwargs = set(kwargs) - set({'n'})
        plural_kwargs = set(kwargs)

        # If using trans methods with `context`, remove it since we are
        # only interested in the singular and plural strings (if any)
        if method_name in ['_p', '_np']:
            args = args[1:]

        # Iterate on strings passed to the trans method. Could be just a
        # singular string or a singular and a plural. We use the index to
        # determine which one is used.
        for idx, arg in enumerate(args):
            found_vars = set()
            check_arg = arg[:]
            check_kwargs = {}
            while True:
                try:
                    check_arg.format(**check_kwargs)
                except KeyError as err:
                    key = err.args[0]
                    found_vars.add(key)
                    check_kwargs[key] = 0
                    continue

                break

            if idx == 0:
                check_1 = singular_kwargs - found_vars
                check_2 = found_vars - singular_kwargs
            else:
                check_1 = plural_kwargs - found_vars
                check_2 = found_vars - plural_kwargs

            if check_1 or check_2:
                error = (arg, check_1.union(check_2))
                self._trans_errors.append(error)

    def visit_Call(self, node):
        method_name, args, kwargs = '', [], []
        with suppress(AttributeError):
            if node.func.value.id == 'trans':
                method_name = node.func.attr
                # Args
                for item in [arg.value for arg in node.args]:
                    args.append(item)
                    self._found.add(item)
                # Kwargs
                kwargs = [
                    kw.arg for kw in node.keywords if kw.arg != 'deferred'
                ]

        if method_name:
            self._check_vars(method_name, args, kwargs)

        self.generic_visit(node)

    def reset(self):
        """Reset variables storing found strings and translation errors."""
        self._found = set()
        self._trans_errors = []


show_trans_strings = FindTransStrings()


def _find_func_definitions(
    node: ast.AST, defs: list[ast.FunctionDef] | None = None
) -> list[ast.FunctionDef]:
    """Find all functions definition recrusively.

    This also find functions nested inside other functions.

    Parameters
    ----------
    node : ast.Node
        The initial node of the ast.
    defs : list of ast.FunctionDef
        A list of function definitions to accumulate.

    Returns
    -------
    list of ast.FunctionDef
        Function definitions found in `node`.
    """
    try:
        body = node.body
    except AttributeError:
        body = []

    if defs is None:
        defs = []

    for node in body:
        _find_func_definitions(node, defs=defs)
        if isinstance(node, ast.FunctionDef):
            defs.append(node)

    return defs


def find_files(
    path: str,
    skip_folders: tuple,
    skip_files: tuple,
    extensions: tuple = ('.py',),
) -> list[str]:
    """Find recursively all files in path.

    Parameters
    ----------
    path : str
        Path to a folder to find files in.
    skip_folders : tuple
        Skip folders containing folder to skip
    skip_files : tuple
        Skip files.
    extensions : tuple, optional
        Extensions to filter by. Default is (".py", )

    Returns
    -------
    list
        Sorted list of found files.
    """
    found_files = []
    for root, _dirs, files in os.walk(path, topdown=False):
        for filename in files:
            fpath = os.path.join(root, filename)

            if any(folder in fpath for folder in skip_folders):
                continue

            if fpath in skip_files:
                continue

            if filename.endswith(extensions):
                found_files.append(fpath)

    return sorted(found_files)


def find_docstrings(fpath: str) -> dict[str, str]:
    """Find all docstrings in file path.

    Parameters
    ----------
    fpath : str
        File path.

    Returns
    -------
    dict
        Simplified string as keys and the value is the original docstring
        found.
    """
    with open(fpath) as fh:
        data = fh.read()

    module = ast.parse(data)
    docstrings = []
    function_definitions = _find_func_definitions(module)
    docstrings.extend([ast.get_docstring(f) for f in function_definitions])
    class_definitions = [
        node for node in module.body if isinstance(node, ast.ClassDef)
    ]
    docstrings.extend([ast.get_docstring(f) for f in class_definitions])
    method_definitions = []

    for class_def in class_definitions:
        method_definitions.extend(
            [
                node
                for node in class_def.body
                if isinstance(node, ast.FunctionDef)
            ]
        )

    docstrings.extend([ast.get_docstring(f) for f in method_definitions])
    docstrings.append(ast.get_docstring(module))
    docstrings = [doc for doc in docstrings if doc]

    results = {}
    for doc in docstrings:
        key = ' '.join([it for it in doc.split() if it != ''])
        results[key] = doc

    return results


def compress_str(gen):
    """
    This function takes a stream of token and tries to join
    consecutive strings.

    This is usefull for long translation string to be broken across
    many lines.

    This should support both joined strings without backslashes:

        trans._(
            "this"
            "will"
            "work"
        )

    Those have NL in between each STING.

    The following will work as well:


        trans._(
            "this"\
            "as"\
            "well"
        )

    Those are just a sequence of STRING


    There _might_ be edge cases with quotes, but I'm unsure

    """
    acc, acc_line = [], None
    for toktype, tokstr, (lineno, _), _, _ in gen:
        if toktype not in (tokenize.STRING, tokenize.NL):
            if acc:
                nt = repr(''.join(acc))
                yield tokenize.STRING, nt, acc_line
                acc, acc_line = [], None
            yield toktype, tokstr, lineno
        elif toktype == tokenize.STRING:
            if tokstr.startswith(("'", '"')):
                acc.append(eval(tokstr))
            else:
                # b"", f"" ... are Strings
                # the prefix can be more than one letter,
                # like rf, rb...
                trailing_quote = tokstr[-1]
                start_quote_index = tokstr.find(trailing_quote)
                prefix = tokstr[:start_quote_index]
                suffix = tokstr[start_quote_index:]
                assert suffix[0] == suffix[-1]
                assert suffix[0] in ('"', "'")
                if 'b' in prefix:
                    print(
                        'not translating bytestring', tokstr, file=sys.stderr
                    )
                    continue
                # we remove the f as we do not want to evaluate the string
                # if it contains variable. IT will crash as it evaluate in
                # the context of this function.
                safe_tokstr = prefix.replace('f', '') + suffix

                acc.append(eval(safe_tokstr))
            if not acc_line:
                acc_line = lineno
        else:
            yield toktype, tokstr, lineno

    if acc:
        nt = repr(''.join(acc))
        yield tokenize.STRING, nt, acc_line


def find_strings(fpath: str) -> dict[tuple[int, str], tuple[int, str]]:
    """Find all strings (and f-strings) for the given file.

    Parameters
    ----------
    fpath : str
        File path.

    Returns
    -------
    dict
        A dict with a tuple for key and a tuple for value. The tuple contains
        the line number and the stripped string. The value containes the line
        number and the original string.
    """
    strings = {}
    with open(fpath) as f:
        for toktype, tokstr, lineno in compress_str(
            tokenize.generate_tokens(f.readline)
        ):
            if toktype == tokenize.STRING:
                try:
                    string = eval(tokstr)
                except Exception:  # noqa BLE001
                    string = eval(tokstr[1:])

                if isinstance(string, str):
                    key = ' '.join([it for it in string.split() if it != ''])
                    strings[(lineno, key)] = (lineno, string)

    return strings


def find_trans_strings(
    fpath: str,
) -> tuple[dict[str, str], list[tuple[str, set[str]]]]:
    """Find all translation strings for the given file.

    Parameters
    ----------
    fpath : str
        File path.

    Returns
    -------
    tuple
        The first item is a dict with a stripped string as key and the
        orginal string for value. The second item is a list of tuples that
        includes errors in translations.
    """
    with open(fpath) as fh:
        data = fh.read()

    module = ast.parse(data)
    trans_strings = {}
    show_trans_strings.visit(module)
    for string in show_trans_strings._found:
        key = ' '.join(list(string.split()))
        trans_strings[key] = string

    errors = list(show_trans_strings._trans_errors)
    show_trans_strings.reset()
    return trans_strings, errors


def import_module_by_path(fpath: str) -> ModuleType | None:
    """Import a module given py a path.

    Parameters
    ----------
    fpath : str
        The path to the file to import as module.

    Returns
    -------
    ModuleType or None
        The imported module or `None`.
    """
    import importlib.util

    fpath = fpath.replace('\\', '/')
    module_name = fpath.replace('.py', '').replace('/', '.')

    try:
        module = importlib.import_module(module_name)
    except ModuleNotFoundError:
        module = None

    return module


def find_issues(
    paths: list[str], skip_words: list[str]
) -> tuple[StringIssuesDict, OutdatedStringsDict, TranslationErrorsDict]:
    """Find strings that have not been translated, and errors in translations.

    This will not raise errors but return a list with found issues wo they
    can be fixed at once.

    Parameters
    ----------
    paths : list of str
        List of paths to files to check.
    skip_words : list of str
        List of words that should be skipped inside the given file.

    Returns
    -------
    tuple
        The first item is a dictionary of the list of issues found per path.
        Each issue is a tuple with line number and the untranslated string.
        The second item is a dictionary of files that contain outdated
        skipped strings. The third item is a dictionary of the translation
        errors found per path. Translation errors referes to missing
        interpolation variables, or spelling errors of the `deferred` keyword.
    """
    issues = {}
    outdated_strings = {}
    trans_errors = {}
    for fpath in paths:
        issues[fpath] = []
        strings = find_strings(fpath)
        trans_strings, errors = find_trans_strings(fpath)
        doc_strings = find_docstrings(fpath)

        skip_words_for_file = skip_words.get(fpath, [])
        skip_words_for_file_check = skip_words_for_file[:]
        module = import_module_by_path(fpath)
        if module is None:
            raise RuntimeError(f'Error loading {fpath}')

        try:
            __all__strings = module.__all__
            if __all__strings is None:
                __all__strings = []
        except AttributeError:
            __all__strings = []

        for key in strings:
            _lineno, string = key
            _lineno, value = strings[key]

            if (
                string not in doc_strings
                and string not in trans_strings
                and value not in skip_words_for_file
                and value not in __all__strings
                and string != ''
                and string.strip() != ''
                and value not in SKIP_WORDS_GLOBAL
            ):
                issues[fpath].append((_lineno, value))
            elif value in skip_words_for_file_check:
                skip_words_for_file_check.remove(value)

        if skip_words_for_file_check:
            outdated_strings[fpath] = skip_words_for_file_check

        if errors:
            trans_errors[fpath] = errors

        if not issues[fpath]:
            issues.pop(fpath)

    return issues, outdated_strings, trans_errors


# --- Fixture
# ----------------------------------------------------------------------------
def _checks():
    paths = find_files(NAPARI_MODULE, SKIP_FOLDERS, SKIP_FILES)
    issues, outdated_strings, trans_errors = find_issues(paths, SKIP_WORDS)
    return issues, outdated_strings, trans_errors


@pytest.fixture(scope='module')
def checks():
    return _checks()


# --- Tests
# ----------------------------------------------------------------------------
def test_missing_translations(checks):
    issues, _, _ = checks
    print(
        '\nSome strings on the following files might need to be translated '
        'or added to the skip list.\nSkip list is located at '
        '`tools/strings_list.py` file.\n\n'
    )
    for fpath, values in issues.items():
        print(f'{fpath}\n{"*" * len(fpath)}')
        unique_values = set()
        for line, value in values:
            unique_values.add(value)
            print(f'{line}:\t{value!r}')

        print('\n')

        if fpath in SKIP_WORDS:
            print(
                f"List below can be copied directly to `tools/strings_list.py` file inside the '{fpath}' key:\n"
            )
            for value in sorted(unique_values):
                print(f'        {value!r},')
        else:
            print(
                'List below can be copied directly to `tools/strings_list.py` file:\n'
            )
            print(f'    {fpath!r}: [')
            for value in sorted(unique_values):
                print(f'        {value!r},')
            print('    ],')

        print('\n')

    no_issues = not issues
    assert no_issues


def test_outdated_string_skips(checks):
    _, outdated_strings, _ = checks
    print(
        '\nSome strings on the skip list on the `tools/strings_list.py` are '
        'outdated.\nPlease remove them from the skip list.\n\n'
    )
    for fpath, values in outdated_strings.items():
        print(f'{fpath}\n{"*" * len(fpath)}')
        print(', '.join(repr(value) for value in values))
        print('')

    no_outdated_strings = not outdated_strings
    assert no_outdated_strings


def test_translation_errors(checks):
    _, _, trans_errors = checks
    print(
        '\nThe following translation strings do not provide some '
        'interpolation variables:\n\n'
    )
    for fpath, errors in trans_errors.items():
        print(f'{fpath}\n{"*" * len(fpath)}')
        for string, variables in errors:
            print(f'String:\t\t{string!r}')
            print(
                f'Variables:\t{", ".join(repr(value) for value in variables)}'
            )
            print('')

        print('')

    no_trans_errors = not trans_errors
    assert no_trans_errors


def getch():
    fd = sys.stdin.fileno()
    old_settings = termios.tcgetattr(fd)
    try:
        tty.setraw(sys.stdin.fileno())
        ch = sys.stdin.read(1)
    finally:
        termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
    return ch


GREEN = '\x1b[1;32m'
RED = '\x1b[1;31m'
NORMAL = '\x1b[1;0m'


def print_colored_diff(old, new):
    lines = list(difflib.unified_diff(old.splitlines(), new.splitlines()))
    for line in lines[2:]:
        if line.startswith('-'):
            print(f'{RED}{line}{NORMAL}')
        elif line.startswith('+'):
            print(f'{GREEN}{line}{NORMAL}')
        else:
            print(line)


def clear_screen():
    print(chr(27) + '[2J')


def _compute_autosugg(raw_code, text):
    raw_code[:]
    start = raw_code.find(f"'{text}'")
    if start == -1:
        start = raw_code.find(f'"{text}"')
    if start == -1:
        return None, False
    stop = start + len(text) + 2
    rawt = raw_code[start:stop]

    sugg = raw_code[:start] + 'trans._(' + rawt + ')' + raw_code[stop:]
    if sugg[start - 1] == 'f':
        return None, False
    return sugg, True


if __name__ == '__main__':
    issues, outdated_strings, trans_errors = _checks()
    import difflib
    import json
    import pathlib

    edit_cmd = sys.argv[1] if len(sys.argv) > 1 else None

    pth = pathlib.Path(__file__).parent / 'string_list.json'
    data = json.loads(pth.read_text())
    for file, items in outdated_strings.items():
        for to_remove in set(items):
            # we don't use set logic to keep the order the same as in the target
            # files.
            data['SKIP_WORDS'][file].remove(to_remove)

    break_ = False

    n_issues = sum([len(m) for m in issues.values()])

    for file, missing in issues.items():
        raw_code = Path(file).read_text()
        code = raw_code.splitlines()
        if break_:
            break
        for line, text in missing:
            # skip current item if it has been added to current list
            # this happens when a new strings is often added many time
            # in the same file.
            if text in data['SKIP_WORDS'].get(file, []):
                continue

            sugg, autosugg = _compute_autosugg(raw_code, text)

            clear_screen()
            print(
                f'{RED}=== About {n_issues} items  in {len(issues)} files to review ==={NORMAL}'
            )

            print()
            print(f'{RED}{file}:{line}{NORMAL}', GREEN, repr(text), NORMAL)
            if autosugg:
                print_colored_diff(raw_code, sugg)
            else:
                print(f'{RED}f-string nedds manual intervention{NORMAL}')
                for lt in code[line - 3 : line - 1]:
                    print(' ', lt)
                print('>', code[line - 1].replace(text, GREEN + text + NORMAL))
                for lt in code[line : line + 3]:
                    print(' ', lt)
            print()

            print()
            print(
                f'{RED}i{NORMAL} : ignore -  add to ignored localised strings'
            )
            print(f'{RED}c{NORMAL} : continue -  go to next')
            if autosugg:
                print(f'{RED}a{NORMAL} : Apply Auto suggestion')
            else:
                print('- : Auto suggestion  not available here')
            if edit_cmd:
                print(f'{RED}e{NORMAL} : EDIT - using {edit_cmd!r}')
            else:
                print(
                    "- : Edit not available, call with python tools/validate_strings.py  '$COMMAND {filename} {linenumber} '"
                )
            print(f'{RED}s{NORMAL} : save and quit')
            print('> ', end='')
            sys.stdout.flush()
            val = getch()
            if val == 'a' and autosugg:
                content = Path(file).read_text()
                new_content, _ = _compute_autosugg(content, text)
                Path(file).write_text(new_content)

            if val == 'e' and edit_cmd:
                subprocess.run(
                    edit_cmd.format(filename=file, linenumber=line).split(' ')
                )
            if val == 'c':
                continue
            if val == 'i':
                data['SKIP_WORDS'].setdefault(file, []).append(text)
            elif val == 'q':
                import sys

                sys.exit(0)
            elif val == 's':
                break_ = True
                break

    pth.write_text(json.dumps(data, indent=4, sort_keys=True))
    # test_outdated_string_skips(issues, outdated_strings, trans_errors)



import numpy as np

from napari.layers.points._points_utils import coerce_symbols


class CoerceSymbolsSuite:
    def setup(self):
        self.symbols1 = np.array(['o' for _ in range(10**6)])
        self.symbols2 = np.array(['o' for _ in range(10**6)])
        self.symbols2[10000] = 's'

    def time_coerce_symbols1(self):
        coerce_symbols(self.symbols1)

    def time_coerce_symbols2(self):
        coerce_symbols(self.symbols2)


if __name__ == '__main__':
    from utils import run_benchmark

    run_benchmark()



import gc
import os
import weakref

import numpy as np
import objgraph
import psutil
import qtpy

import napari

process = psutil.Process(os.getpid())
viewer = napari.Viewer()

print('mem', process.memory_info().rss)

for _ in range(0):
    print(viewer.add_image(np.random.random((60, 1000, 1000))).name)
for _ in range(2):
    print(viewer.add_labels((np.random.random((2, 1000, 1000)) * 10).astype(np.uint8)).name)

print('mem', process.memory_info().rss)

# napari.run()

print('controls', viewer.window.qt_viewer.controls.widgets)
li = weakref.ref(viewer.layers[0])
data_li = weakref.ref(li()._data)
controls = weakref.ref(viewer.window.qt_viewer.controls.widgets[li()])
objgraph.show_backrefs(li(), filename='base.png')
del viewer.layers[0]
qtpy.QtGui.QGuiApplication.processEvents()
gc.collect()
gc.collect()
print(li())
objgraph.show_backrefs(li(), max_depth=10, filename='test.png', refcounts=True)
objgraph.show_backrefs(controls(), max_depth=10, filename='controls.png', refcounts=True)
objgraph.show_backrefs(data_li(), max_depth=10,  filename='test_data.png')
print('controls', viewer.window.qt_viewer.controls.widgets)
print('controls', gc.get_referrers(controls()))
print('controls', controls().parent())
#print("controls", controls().parent().indexOf(controls()))
print(gc.get_referrers(li()))
print(gc.get_referrers(li())[1])
print(gc.get_referrers(gc.get_referrers(gc.get_referrers(li())[0])))
res = gc.get_referrers(gc.get_referrers(gc.get_referrers(li())[0])[0])
print(res)
#print(type(res[0]))



# See "Writing benchmarks" in the asv docs for more information.
# https://asv.readthedocs.io/en/latest/writing_benchmarks.html
# or the napari documentation on benchmarking
# https://github.com/napari/napari/blob/main/docs/BENCHMARKS.md
import os
from functools import lru_cache
from itertools import cycle

import numpy as np
from packaging.version import parse as parse_version
from qtpy.QtWidgets import QApplication
from skimage.morphology import diamond, octahedron
from vispy.app import MouseEvent

import napari
import napari.layers
from napari.components.viewer_model import ViewerModel
from napari.qt import QtViewer
from napari.utils.colormaps import DirectLabelColormap

from .utils import Skip

NAPARI_0_4_19 = parse_version(napari.__version__) <= parse_version('0.4.19')


class QtViewerSingleLabelsSuite:
    """Benchmarks for editing a single labels layer in the viewer."""

    data: np.ndarray
    viewer: napari.Viewer
    layer: napari.layers.Labels
    event: MouseEvent

    def setup(self):
        _ = QApplication.instance() or QApplication([])
        rng = np.random.default_rng(0)
        self.data = rng.integers(10, size=(512, 512))
        self.viewer = napari.Viewer()
        self.layer = self.viewer.add_labels(self.data)
        self.layer.brush_size = 10
        self.layer.mode = 'paint'
        self.layer.selected_label = 3
        self.event = MouseEvent(
            type='mouse_move',
            press_event=MouseEvent(type='mouse_press', pos=(505, 500)),
            pos=(500, 500),
            view_direction=None,
        )
        if NAPARI_0_4_19:
            self.camera = self.viewer.window._qt_viewer.view.camera
        else:
            self.camera = self.viewer.window._qt_viewer.canvas.view.camera

    def teardown(self):
        self.viewer.window.close()

    def time_zoom(self):
        """Time to zoom in and zoom out."""
        self.camera.zoom(0.5, center=(0.5, 0.5))
        self.camera.zoom(2.0, center=(0.5, 0.5))

    def time_set_view_slice(self):
        """Time to set view slice."""
        if hasattr(self.layer, '_slicing_state'):
            self.layer._slicing_state._set_view_slice()
        else:
            # before https://github.com/napari/napari/pull/8254
            self.layer._set_view_slice()

    def time_refresh(self):
        """Time to refresh view."""
        self.layer.refresh()

    def time_update_thumbnail(self):
        """Time to update thumbnail."""
        self.layer._update_thumbnail()

    def time_get_value(self):
        """Time to get current value."""
        self.layer.get_value((0,) * 2)

    def time_raw_to_displayed(self):
        """Time to convert raw to displayed."""
        self.layer._raw_to_displayed(self.layer._slice.image.raw)

    def time_paint(self):
        """Time to paint."""
        self.layer.paint((0,) * 2, self.layer.selected_label)

    def time_fill(self):
        """Time to fill."""
        self.layer.fill(
            (0,) * 2,
            1,
            self.layer.selected_label,
        )

    def time_on_mouse_move(self):
        """Time to drag paint on mouse move."""
        if NAPARI_0_4_19:
            self.viewer.window._qt_viewer.on_mouse_move(self.event)
        else:
            self.viewer.window._qt_viewer.canvas._on_mouse_move(self.event)


@lru_cache
def setup_rendering_data(radius, dtype):
    if radius < 1000:
        data = octahedron(radius=radius, dtype=dtype)
    else:
        data = np.zeros((radius // 50, radius * 2, radius * 2), dtype=dtype)
        for i in range(1, data.shape[0] // 2):
            part = diamond(radius=i * 100, dtype=dtype)
            shift = (data.shape[1] - part.shape[0]) // 2
            data[i, shift : -shift - 1, shift : -shift - 1] = part
            data[-i - 1, shift : -shift - 1, shift : -shift - 1] = part

    count = np.count_nonzero(data)
    data[data > 0] = np.random.randint(
        1, min(2000, np.iinfo(dtype).max), size=count, dtype=dtype
    )

    return data


class LabelRendering:
    """Benchmarks for rendering the Labels layer."""

    param_names = ['radius', 'dtype', 'mode']
    params = (
        [10, 30, 300, 1500],
        [np.uint8, np.uint16, np.uint32],
        ['auto', 'direct'],
    )
    skip_params = Skip(
        if_in_pr=lambda radius, *_: radius > 20,
        if_on_ci=lambda radius, *_: radius > 20,
    )

    def setup(self, radius, dtype, label_mode):
        self.steps = 4 if 'GITHUB_ACTIONS' in os.environ else 10
        self.app = QApplication.instance() or QApplication([])
        self.data = setup_rendering_data(radius, dtype)
        scale = self.data.shape[-1] / np.array(self.data.shape)
        self.viewer = ViewerModel()
        self.qt_viewr = QtViewer(self.viewer)
        self.layer = self.viewer.add_labels(self.data, scale=scale)
        if label_mode == 'direct':
            colors = dict(
                zip(
                    range(10, 2000),
                    cycle(['red', 'green', 'blue', 'pink', 'magenta']),
                )
            )
            colors[None] = 'yellow'
            colors[0] = 'transparent'
            self.layer.colormap = DirectLabelColormap(color_dict=colors)
        self.qt_viewr.show()

    @staticmethod
    def teardown(self, *_):
        if hasattr(self, 'viewer'):
            self.qt_viewr.close()

    def _time_iterate_components(self, *_):
        """Time to iterate over components."""
        self.layer.show_selected_label = True
        for i in range(0, 201, (200 // self.steps) or 1):
            self.layer.selected_label = i
            self.app.processEvents()

    def _time_zoom_change(self, *_):
        """Time to zoom in and zoom out."""
        initial_zoom = self.viewer.camera.zoom
        self.viewer.camera.zoom = 0.5 * initial_zoom
        self.app.processEvents()
        self.viewer.camera.zoom = 2 * initial_zoom
        self.app.processEvents()


class LabelRenderingSuite2D(LabelRendering):
    def setup(self, radius, dtype, label_mode):
        super().setup(radius, dtype, label_mode)
        self.viewer.dims.ndisplay = 2
        self.app.processEvents()

    def time_iterate_over_z(self, *_):
        """Time to render the layer."""
        z_size = self.data.shape[0]
        for i in range(0, z_size, z_size // (self.steps * 2)):
            self.viewer.dims.set_point(0, i)
            self.app.processEvents()

    def time_load_3d(self, *_):
        """Time to first render of the layer in 3D."""
        self.app.processEvents()
        self.viewer.dims.ndisplay = 3
        self.app.processEvents()
        self.viewer.dims.ndisplay = 2
        self.app.processEvents()

    def time_iterate_components(self, *args):
        self._time_iterate_components(*args)

    def time_zoom_change(self, *args):
        self._time_zoom_change(*args)


class LabelRenderingSuite3D(LabelRendering):
    def setup(self, radius, dtype, label_mode):
        super().setup(radius, dtype, label_mode)
        self.viewer.dims.ndisplay = 3
        self.app.processEvents()

    def time_rotate(self, *_):
        """Time to rotate the layer."""
        for i in range(0, (self.steps * 20), 5):
            self.viewer.camera.angles = (0, i / 2, i)
            self.app.processEvents()

    def time_iterate_components(self, *args):
        self._time_iterate_components(*args)

    def time_zoom_change(self, *args):
        self._time_zoom_change(*args)


if __name__ == '__main__':
    from utils import run_benchmark

    run_benchmark()



from collections.abc import Generator

import pytest
from pytestqt.qtbot import QtBot

from napari._qt.widgets.qt_dims import QtDims
from napari.components import Dims


@pytest.fixture
def qt_dims(qtbot: QtBot) -> Generator[QtDims, None, None]:
    """Fixture that provides a QtDims and ensures
    that animation is stopped after the test failed.

    The test itself should ensure that the animation is stopped,
    but in case of a failure before the end of the test,
    this fixture will take care of stopping it
    to avoid segfaults in following tests.
    """
    dims = Dims(ndim=2)
    qt_dims = QtDims(dims)
    qtbot.addWidget(qt_dims)
    yield qt_dims
    if qt_dims.is_playing:
        qt_dims.stop()
        qtbot.waitUntil(lambda: not qt_dims.is_playing)



"""
Minimum blending
================

Demonstrates how to use the `minimum` blending mode with inverted colormaps.
`minimum` blending uses the minimum value of each R, G, B channel for each pixel.
`minimum` blending can be used to yield multichannel color images on a white
background, when the channels have inverted colormaps assigned.
An inverted colormap is one where white [1, 1, 1] is used to represent the lowest
values, as opposed to the more conventional black [0, 0, 0]. For example, try the
colormaps prefixed with *I*, such as *I Forest* or *I Bordeaux*, from
ChrisLUTs: https://github.com/cleterrier/ChrisLUTs .

.. tags:: visualization-basic
"""

from skimage import data

import napari

# create a viewer
viewer = napari.Viewer()

# Add the cells3d example image, using the two inverted colormaps
# and minimum blending mode. Note that the bottom-most layer
# must be translucent or opaque to prevent blending with the canvas.
viewer.add_image(data.cells3d(),
                            name=['membrane', 'nuclei'],
                            channel_axis=1,
                            contrast_limits = [[1110, 23855], [1600, 50000]],
                            colormap = ['I Purple', 'I Orange'],
                            blending= ['translucent_no_depth', 'minimum']
                            )

if __name__ == '__main__':
    napari.run()



# See "Writing benchmarks" in the asv docs for more information.
# https://asv.readthedocs.io/en/latest/writing_benchmarks.html
# or the napari documentation on benchmarking
# https://github.com/napari/napari/blob/main/docs/BENCHMARKS.md
from copy import copy

import numpy as np
from packaging.version import parse as parse_version

import napari
from napari.components.dims import Dims
from napari.layers import Labels
from napari.utils.colormaps import DirectLabelColormap

from .utils import Skip, labeled_particles

MAX_VAL = 2**23

NAPARI_0_4_19 = parse_version(napari.__version__) <= parse_version('0.4.19')


class Labels2DSuite:
    """Benchmarks for the Labels layer with 2D data"""

    param_names = ['n', 'dtype']
    params = ([2**i for i in range(4, 13)], [np.uint8, np.int32])

    skip_params = Skip(if_in_pr=lambda n, dtype: n > 2**5)

    def setup(self, n, dtype):
        np.random.seed(0)
        self.data = labeled_particles(
            (n, n), dtype=dtype, n=int(np.log2(n) ** 2), seed=1
        )
        self.layer = Labels(self.data)
        self.layer._raw_to_displayed(self.data, (slice(0, n), slice(0, n)))

    def time_create_layer(self, *_):
        """Time to create layer."""
        Labels(self.data)

    def time_set_view_slice(self, *_):
        """Time to set view slice."""
        if hasattr(self.layer, '_slicing_state'):
            self.layer._slicing_state._set_view_slice()
        else:
            # before https://github.com/napari/napari/pull/8254
            self.layer._set_view_slice()

    def time_refresh(self, *_):
        """Time to refresh view."""
        self.layer.refresh()

    def time_update_thumbnail(self, *_):
        """Time to update thumbnail."""
        self.layer._update_thumbnail()

    def time_get_value(self, *_):
        """Time to get current value."""
        self.layer.get_value((0,) * 2)

    def time_raw_to_displayed(self, *_):
        """Time to convert raw to displayed."""
        self.layer._slice.image.raw[0, :] += 1  # simulate changes
        self.layer._raw_to_displayed(self.layer._slice.image.raw)

    def time_paint_circle(self, *_):
        """Time to paint circle."""
        self.layer.paint((0,) * 2, self.layer.selected_label)

    def time_fill(self, *_):
        """Time to fill."""
        self.layer.fill(
            (0,) * 2,
            1,
            self.layer.selected_label,
        )

    def mem_layer(self, *_):
        """Memory used by layer."""
        return copy(self.layer)

    def mem_data(self, *_):
        """Memory used by raw data."""
        return self.data


class LabelsDrawing2DSuite:
    """Benchmark for brush drawing in the Labels layer with 2D data."""

    param_names = ['n', 'brush_size', 'color_mode', 'contour']
    params = ([512, 3072], [8, 64, 256], ['auto', 'direct'], [0, 1])
    skip_params = Skip(
        if_in_pr=lambda n, brush_size, *_: n > 512 or brush_size > 64
    )

    def setup(self, n, brush_size, color_mode, contour):
        np.random.seed(0)
        self.data = labeled_particles(
            (n, n), dtype=np.int32, n=int(np.log2(n) ** 2), seed=1
        )

        self.layer = Labels(self.data)

        if color_mode == 'direct':
            random_label_ids = np.random.randint(64, size=50)
            colors = {i + 1: np.random.random(4) for i in random_label_ids}
            colors[None] = np.array([0, 0, 0, 0.3])
            self.layer.colormap = DirectLabelColormap(color_dict=colors)

        self.layer.brush_size = brush_size
        self.layer.contour = contour
        self.layer.mode = 'paint'

    def time_draw(self, n, brush_size, color_mode, contour):
        new_label = self.layer._slice.image.raw[0, 0] + 1

        with self.layer.block_history():
            last_coord = (0, 0)
            for x in np.linspace(0, n - 1, num=30)[1:]:
                self.layer._draw(
                    new_label=new_label,
                    last_cursor_coord=last_coord,
                    coordinates=(x, x),
                )
                last_coord = (x, x)


class Labels2DColorDirectSuite(Labels2DSuite):
    skip_params = Skip(if_in_pr=lambda n, dtype: n > 32)

    def setup(self, n, dtype):
        np.random.seed(0)
        info = np.iinfo(dtype)
        self.data = labeled_particles(
            (n, n), dtype=dtype, n=int(np.log2(n) ** 2), seed=1
        )
        random_label_ids = np.random.randint(
            low=max(-10000, info.min), high=min(10000, info.max), size=20
        )
        colors = {i + 1: np.random.random(4) for i in random_label_ids}
        colors[None] = np.array([0, 0, 0, 0.3])
        self.layer = Labels(
            self.data, colormap=DirectLabelColormap(color_dict=colors)
        )
        self.layer._raw_to_displayed(
            self.layer._slice.image.raw, (slice(0, n), slice(0, n))
        )


class Labels3DSuite:
    """Benchmarks for the Labels layer with 3D data."""

    param_names = ['n', 'dtype']
    params = ([2**i for i in range(4, 11)], [np.uint8, np.uint32])

    skip_params = Skip(
        if_in_pr=lambda n, dtype: n > 2**6, if_on_ci=lambda n, dtype: n > 2**9
    )
    # CI skip above 2**9 because of memory limits

    def setup(self, n, dtype):
        np.random.seed(0)
        self.data = labeled_particles(
            (n, n, n), dtype=dtype, n=int(np.log2(n) ** 2), seed=1
        )
        self.layer = Labels(self.data)
        if NAPARI_0_4_19:
            self.layer._slice_dims((0, 0, 0), ndisplay=3)
        else:
            self.layer._slice_dims(Dims(ndim=3, ndisplay=3))
        self.layer._raw_to_displayed(
            self.layer._slice.image.raw,
            (slice(0, n), slice(0, n), slice(0, n)),
        )

    # @mark.skip_params_if([(2**i,) for i in range(6, 11)], condition="PR" in os.environ)
    def time_create_layer(self, *_):
        """Time to create layer."""
        Labels(self.data)

    def time_set_view_slice(self, *_):
        """Time to set view slice."""
        if hasattr(self.layer, '_slicing_state'):
            self.layer._slicing_state._set_view_slice()
        else:
            # before https://github.com/napari/napari/pull/8254
            self.layer._set_view_slice()

    def time_refresh(self, *_):
        """Time to refresh view."""
        self.layer.refresh()

    def time_update_thumbnail(self, *_):
        """Time to update thumbnail."""
        self.layer._update_thumbnail()

    def time_get_value(self, *_):
        """Time to get current value."""
        self.layer.get_value((0,) * 3)

    def time_raw_to_displayed(self, *_):
        """Time to convert raw to displayed."""
        self.layer._slice.image.raw[0, 0, :] += 1  # simulate changes
        self.layer._raw_to_displayed(self.layer._slice.image.raw)

    def time_paint_circle(self, *_):
        """Time to paint circle."""
        self.layer.paint((0,) * 3, self.layer.selected_label)

    def time_fill(self, *_):
        """Time to fill."""
        self.layer.fill(
            (0,) * 3,
            1,
            self.layer.selected_label,
        )

    def mem_layer(self, *_):
        """Memory used by layer."""
        return copy(self.layer)

    def mem_data(self, *_):
        """Memory used by raw data."""
        return self.data


if __name__ == '__main__':
    from utils import run_benchmark

    run_benchmark()



"""
napari without gui_qt
=====================

Alternative to using napari.gui_qt() context manager.

This is here for historical purposes, to the transition away from
the "gui_qt()" context manager.

.. tags:: historical
"""

from collections import Counter

from skimage import data

import napari

viewer = napari.Viewer()
layer = viewer.add_image(data.astronaut(), rgb=True)

# You can do anything you would normally do with the viewer object
# like take a
screenshot = viewer.screenshot()

print('Maximum value', screenshot.max())

# To see the napari viewer and interact with the graphical user interface,
# use `napari.run()`.  (it's similar to `plt.show` in matplotlib)
# If you only wanted the screenshot then you could skip this entirely.
# *run* will *block execution of your script* until the window is closed.
if __name__ == '__main__':
    napari.run()

    # When the window is closed, your script continues and you can still inspect
    # the viewer object.  For example, add click the buttons to add various layer
    # types when the window is open and see the result below:

    print('Your viewer has the following layers:')
    for name, n in Counter(type(x).__name__ for x in viewer.layers).most_common():
        print(f'   {name:<7}: {n}')



"""
Interactive move rectangle
==========================

Shift a rectangle along its normal vector in 3D

.. tags:: experimental
"""

import numpy as np

import napari

rectangle = np.array(
    [
        [50, 75, 75],
        [50, 125, 75],
        [100, 125, 125],
        [100, 75, 125]
    ],
    dtype=float
)

shapes_data = np.array(rectangle)
normal_vector = np.cross(
    rectangle[0] - rectangle[1], rectangle[2] - rectangle[1]
)
normal_vector /= np.linalg.norm(normal_vector)

viewer = napari.Viewer(ndisplay=3)

shapes_layer = viewer.add_shapes(
    data=shapes_data,
    face_color='blue'
)
viewer.camera.angles = (-170, -20, -170)
viewer.camera.zoom = 1.5
viewer.text_overlay.visible = True
viewer.text_overlay.text = """'click and drag the rectangle to create copies along its normal vector
"""


@shapes_layer.mouse_drag_callbacks.append
def move_rectangle_along_normal(layer, event):
    shape_index, _ = layer.get_value(
        position=event.position,
        view_direction=event.view_direction,
        dims_displayed=event.dims_displayed
    )
    if shape_index is None:
        return

    layer.mouse_pan = False

    start_position = np.copy(event.position)
    yield
    while event.type == 'mouse_move':
        projected_distance = layer.projected_distance_from_mouse_drag(
            start_position=start_position,
            end_position=event.position,
            view_direction=event.view_direction,
            vector=normal_vector,
            dims_displayed=event.dims_displayed,
        )
        shift_data_coordinates = projected_distance * normal_vector
        new_rectangle = layer.data[shape_index] + shift_data_coordinates
        layer.add(new_rectangle)
        yield
    layer.mouse_pan = True


if __name__ == '__main__':
    napari.run()



"""
3D Layer Bounding Box Overlay
=============================

Display multiple layer types in 3D and add a bounding box overlay to each layer.

The bounding box overlay is a visual representation of the extents of each layer.
Image and label layers have data extents that are the same as the image data.
Shapes and points layers have data extents that represent the convex hull of the data.

For an example showing how to modify the properties of the bounding box overlay,
see :ref:`sphx_glr_gallery_image_border.py`.

.. tags:: visualization-advanced, visualization-nD
"""

import numpy as np
from skimage import data, feature, filters, morphology

import napari

# get sample 3D image data, threshold, and find maxima
cells3d = data.cells3d()
nuclei = cells3d[:, 1]

nuclei_smoothed = filters.gaussian(nuclei, sigma=5)
nuclei_thresholded = nuclei_smoothed > filters.threshold_otsu(nuclei_smoothed)
nuclei_labels = morphology.label(nuclei_thresholded)
nuclei_points = feature.peak_local_max(nuclei_smoothed, min_distance=20)

# create an arbitrary path for display
path = np.array([
    [0, 10, 10],
    [20, 5, 15],
    [56, 70, 21],
    [127, 127, 127]
])

# create the viewer and display the different layer types
viewer = napari.Viewer()
viewer.add_image(nuclei, name='nuclei', contrast_limits = (10000, 65355))
viewer.add_labels(nuclei_labels, name='nuclei labels')
viewer.add_points(
    nuclei_points, name='nuclei maxima', blending='additive', opacity=0.5
)
viewer.add_shapes(
    path, name='path', shape_type='path', blending='additive', edge_color='yellow'
)

# add a bounding box overlay to each layer, then change the color
for layer in viewer.layers:
    layer.bounding_box.visible = True

viewer.layers['nuclei labels'].bounding_box.line_color = 'cyan'
viewer.layers['nuclei maxima'].bounding_box.line_color = 'orange'
viewer.layers['path'].bounding_box.line_color = 'magenta'

# set the view to 3D and rotate camera
viewer.dims.ndisplay = 3
viewer.camera.angles = (2, 15, 150)

if __name__ == '__main__':
    napari.run()



"""
Affine Transforms 3D
====================

Translate and rotate an object in napari using an affine transformation. Affine for 3D objects is represented as a 4x4 matrix,
while affine for 2D objects is represented as a 3x3 matrix. For 3D objects the 4 columns represent the rotate, scale, shear, and
translate respectively, and by default, affine is set as a 4x4 identity matrix. By modifying the respective values in this matrix,
which is represented as a numpy array, we can move and adjust objects in the napari viewer.

.. tags:: visualization-advanced
"""

import meshio
import numpy as np
import pooch
from magicgui import magicgui

import napari

cup_path = pooch.retrieve(
    url="https://raw.githubusercontent.com/8bitbiscuit/cup_and_creamer/main/cup.stl",
    known_hash='b6162298f86c94ebb276eb5f0544409996a8ee81fff9627a2bd4b71309835014',
)
creamer_path = pooch.retrieve(
    url="https://raw.githubusercontent.com/8bitbiscuit/cup_and_creamer/main/creamer.stl",
    known_hash='122256d942694cf40fc1a186a346cda0b31a03fbdec61ea8a351c915f3d0d6ed',
)

# Use meshio to read in stl files
cup_mesh = meshio.read(cup_path)
creamer_mesh = meshio.read(creamer_path)

# Create tuples using the data read in from meshio to be viewed by napari as surfaces
cup_data = (cup_mesh.points, cup_mesh.cells[0].data)
creamer_data = (creamer_mesh.points, creamer_mesh.cells[0].data)

viewer = napari.Viewer()

# Define an original affine transformation to be position of surface upon being loaded into viewer
# Otherwise both objects would be loaded in on the origin in the viewer by default and be on top of one another
og_affine = np.array(
    [
        [1.0, 0.0, 0.0, 0.0],
        [0.0, 1.0, 0.0, 0.0],
        [0.0, 0.0, 1.0, -0.1],
        [0.0, 0.0, 0.0, 1.0],
    ]
)

# Add in the two surfaces we read using meshio
viewer.add_surface(cup_data, name="cup")
viewer.add_surface(
    creamer_data, affine=og_affine, opacity=0.5, blending="additive", name="creamer"
)


@magicgui(call_button='POUR!!')
def pour_creamer():
    # Define a new affine transform for the creamer surface, where it is moved up and at an angle to be 'poured' into the cup
    new_affine = np.array(
        [
            [1.0, 0.0, 0.0, 0.0],
            [0.0, 0.23, -0.97, 0.15],
            [0.0, 0.97, 0.23, -0.08],
            [0.0, 0.0, 0.0, 1.0],
        ]
    )

    # Add a transformed surface with the new affine transform
    # This surface is identical to the other creamer object, save for the new affine transformation
    viewer.add_surface(
        creamer_data, affine=new_affine, opacity=0.5, blending="additive", name="pour"
    )


# Set viewer camera angles and zoom
viewer.dims.ndisplay = 3
viewer.camera.angles = (180, 40, 92)
viewer.camera.zoom = 1972

# Add button to create new layer with transformed object using magicgui
viewer.window.add_dock_widget(pour_creamer)

pour_creamer()

if __name__ == "__main__":
    napari.run()



"""
Live tiffs generator
====================

Simulation of microscope acquisition. This code generates time series tiffs in
an output directory (must be supplied by the user).

.. tags:: experimental
"""


import argparse
import os
import sys
import time

import numpy as np
import tifffile
from skimage import data

parser = argparse.ArgumentParser()
parser.add_argument('outdir', help='output directory for tiffs')
parser.add_argument(
    '--sleep-time',
    help='how long to sleep between volumes, in seconds',
    type=float,
    default=1.0,
)
parser.add_argument(
    '-n', help='total number of volumes', type=int, default=100
)


def main(argv=sys.argv[1:]):
    args = parser.parse_args(argv)
    outdir = args.outdir
    sleep_time = args.sleep_time
    n = args.n
    fractions = np.linspace(0.05, 0.5, n)
    os.makedirs(outdir, exist_ok=True)
    for i, f in enumerate(fractions):
        # We are using skimage binary_blobs which generates synthetic binary
        # image with several rounded blob-like objects and write them into files.
        curr_vol = 255 * data.binary_blobs(
            length=128, blob_size_fraction=0.05, n_dim=3, volume_fraction=f
        ).astype(np.uint8)
        tifffile.imwrite(
            os.path.join(outdir, f'{i}.tiff'), curr_vol, compress=6
        )
        time.sleep(sleep_time)
    # create a final.log file as an indicator for end of acquisition
    with open(os.path.join(outdir, 'final.log'), 'w'):
        pass


if __name__ == '__main__':
    main()



"""
Mouse drag callback
===================

This example shows how to use a mouse drag callback in Napari to compute
and print the average intensity along line shapes as they are moved.

As you drag lines over the image, the script calculates the mean of the
intensity profile along each line and prints the results to the terminal in real time.

.. tags:: gui
"""

import numpy as np
from skimage import data, measure

import napari


def profile_lines(image, shape_layer):
    profile_data = [
        measure.profile_line(image, line[0], line[1], mode='reflect').mean()
        for line in shape_layer.data
    ]
    print(f"profile means: [{', '.join(f'{d:.2f}' for d in profile_data)}]")


np.random.seed(1)
viewer = napari.Viewer()
blobs = data.binary_blobs(length=512, volume_fraction=0.1, n_dim=2)
viewer.add_image(blobs, name='blobs')
line1 = np.array([[11, 13], [111, 113]])
line2 = np.array([[200, 200], [400, 300]])
lines = [line1, line2]
shapes_layer = viewer.add_shapes(
    lines,
    shape_type='line',
    edge_width=5,
    edge_color='coral',
    face_color='royalblue',
)
shapes_layer.mode = 'select'


@shapes_layer.mouse_drag_callbacks.append
def profile_lines_drag(layer, event):
    profile_lines(blobs, layer)
    yield
    while event.type == 'mouse_move':
        profile_lines(blobs, layer)
        # the yield statement allows the mouse UI to keep working while
        # this loop is executed repeatedly
        yield


if __name__ == '__main__':
    napari.run()



from typing import TYPE_CHECKING

from napari._qt.layer_controls.qt_image_controls_base import (
    QtBaseImageControls,
)
from napari._qt.layer_controls.widgets._surface import QtShadingComboBoxControl

if TYPE_CHECKING:
    import napari.layers


class QtSurfaceControls(QtBaseImageControls):
    """Qt view and controls for the napari Surface layer.

    Parameters
    ----------
    layer : napari.layers.Surface
        An instance of a napari Surface layer.

    Attributes
    ----------
    _shading_combobox_control : napari._qt.layer_controls.widgets._surface.QtShadingComboBoxControl
        Widget that wraps comboBox controlling current shading value of the layer.
    """

    layer: 'napari.layers.Surface'
    PAN_ZOOM_ACTION_NAME = 'activate_surface_pan_zoom_mode'
    TRANSFORM_ACTION_NAME = 'activate_surface_transform_mode'

    def __init__(self, layer) -> None:
        super().__init__(layer)

        # Setup widgets controls
        self._shading_combobox_control = QtShadingComboBoxControl(self, layer)
        self._add_widget_controls(self._shading_combobox_control)



#!/usr/bin/env python3
"""
This script reads your project's pyproject.toml (PEP 621 format) and iterates over all
dependencies (both main and extras). For each dependency that is a binary dependency
(i.e. one that provides platform-specific wheels), it checks whether the version specified
as the minimal (from a lower-bound like ">=1.3.0") provides a wheel compatible with a given
Python version (e.g. 3.10).

If not, it searches (in ascending order) for the lowest version (above the specified minimal)
that provides a compatible wheel and reports the problematic packages with the recommended version bump.

Requirements:
  - Python 3.11+ uses the built-in tomllib; older versions require the external 'toml' package.
  - Install requests and packaging (and toml if needed).

Usage: Run this script in your project root (where pyproject.toml is located).

It is chatGPT generated code.
"""

import re
import sys

import requests

# Use built-in tomllib if available; else fallback to external toml
try:
    import tomllib
except ModuleNotFoundError:
    import toml as tomllib

from packaging.requirements import InvalidRequirement, Requirement
from packaging.specifiers import SpecifierSet
from packaging.utils import InvalidWheelFilename, parse_wheel_filename
from packaging.version import InvalidVersion, Version

# --- Helper Functions ---


def parse_min_python_version(requires_python: str):
    """
    Extracts the minimal Python version from a requires-python string (e.g. ">=3.10").
    Assumes a specifier like ">=X.Y" is present.
    """
    m = re.search(r'>=\s*([\d\.]+)', requires_python)
    if m:
        ver_str = m.group(1)
        try:
            return Version(ver_str)
        except InvalidVersion:
            print(f'Could not parse version from requires-python: {ver_str}')
            return None
    else:
        print("No '>=' specifier found in requires-python.")
        return None


def get_specified_min_version(specifier: SpecifierSet) -> Version:
    """
    Returns the effective minimal version from a dependency's specifier set.
    This is the maximum among all lower-bound constraints (">=" or "==").
    """
    lower_bounds = []
    for spec in specifier:
        if spec.operator in ('>=', '=='):
            try:
                lower_bounds.append(Version(spec.version))
            except InvalidVersion:
                continue
    if not lower_bounds:
        return None
    return max(lower_bounds)


def wheel_supports_python(filename: str, required_py_ver: Version) -> bool:
    """
    Determines if the wheel (given by its filename) supports the required Python version.

    Heuristic:
      - Universal wheels (py3-none-any) are accepted.
      - For CPython wheels:
          • If the ABI tag contains "abi3" (e.g. cp38-abi3), assume it works for any CPython version
            that is equal to or newer than the tagged version.
          • Otherwise, the wheel's tag (e.g. "cp310") must match the required version exactly.
    """
    try:
        _, _, _, tags = parse_wheel_filename(filename)
    except InvalidWheelFilename as e:
        print(f'Error parsing wheel filename {filename}: {e}')
        return False

    for tag in tags:
        interpreter = tag.interpreter.lower()
        abi = tag.abi.lower()
        platform = tag.platform.lower()

        # Accept universal pure-Python wheels.
        if interpreter == 'py3' and abi == 'none' and platform == 'any':
            return True

        if interpreter.startswith('cp'):
            ver_str = interpreter[2:]  # e.g. "310" or "39"
            try:
                if len(ver_str) == 2:
                    wheel_major = int(ver_str[0])
                    wheel_minor = int(ver_str[1])
                elif len(ver_str) >= 3:
                    wheel_major = int(ver_str[0])
                    wheel_minor = int(ver_str[1:])
                else:
                    continue
            except ValueError:
                continue

            req_major = required_py_ver.major
            req_minor = required_py_ver.minor

            if 'abi3' in abi:
                # Assume compatibility if the required version is >= the tagged version.
                if (
                    req_major == wheel_major and req_minor >= wheel_minor
                ) or req_major > wheel_major:
                    return True
            else:
                if req_major == wheel_major and req_minor == wheel_minor:
                    return True
    return False


def has_compatible_wheel(files, required_py_ver: Version) -> bool:
    """
    Returns True if any wheel file in the given list is compatible with the required Python version.
    """
    for file_info in files:
        if file_info.get('packagetype') == 'bdist_wheel':
            fname = file_info.get('filename')
            if fname and wheel_supports_python(fname, required_py_ver):
                return True
    return False


def check_version_for_compatible_wheel(
    pkg_name: str, version: Version, required_py_ver: Version
) -> bool:
    """
    Checks whether the given version of a package provides a wheel compatible with the required Python version.
    """
    url = f'https://pypi.org/pypi/{pkg_name}/{version}/json'
    resp = requests.get(url)
    if resp.status_code != 200:
        print(
            f'  Could not fetch data for {pkg_name} version {version} (status {resp.status_code}).'
        )
        return False
    data = resp.json()
    files = data.get('urls', [])
    return has_compatible_wheel(files, required_py_ver)


def find_lowest_version_with_compatible_wheel(
    pkg_name: str,
    spec: SpecifierSet,
    pypi_data: dict,
    required_py_ver: Version,
    current_min: Version,
):
    """
    Among all candidate versions (satisfying the specifier and greater than current_min),
    returns the lowest version that provides a compatible wheel for the required Python version.
    Returns None if none found.
    """
    candidates = []
    releases = pypi_data.get('releases', {})
    for ver_str in releases:
        try:
            ver = Version(ver_str)
        except InvalidVersion:
            continue
        if ver in spec and ver > current_min:
            candidates.append(ver)
    if not candidates:
        return None
    candidates.sort()  # ascending order
    for ver in candidates:
        files = releases.get(str(ver), [])
        if has_compatible_wheel(files, required_py_ver):
            return ver
    return None


def process_dependency(dep_str: str, required_py_ver: Version, origin: str):
    """
    Processes a single dependency string (from main dependencies or extras):
      - Determines the specified minimal version.
      - Checks if that version provides a compatible wheel for the required Python version.
      - If not, searches for the lowest version above the specified minimal version that does.
    Returns a tuple (pkg_name, specified_min, alternative_version) if an issue is found; otherwise None.
    """
    try:
        req_obj = Requirement(dep_str)
    except InvalidRequirement as e:
        print(f"  Failed to parse dependency '{dep_str}': {e}")
        return None

    pkg_name = req_obj.name
    spec = req_obj.specifier or SpecifierSet()
    specified_min = get_specified_min_version(spec)
    if specified_min is None:
        print(
            f'  {pkg_name} ({origin}) has no explicit lower-bound; skipping check.'
        )
        return None

    print(
        f'  {pkg_name} ({origin}): specified minimal version = {specified_min}'
    )

    if check_version_for_compatible_wheel(
        pkg_name, specified_min, required_py_ver
    ):
        print(
            f'    OK: Version {specified_min} provides a compatible wheel for Python {required_py_ver}'
        )
        return None
    # Query PyPI to look for the lowest version above the specified minimal that provides a compatible wheel.
    url = f'https://pypi.org/pypi/{pkg_name}/json'
    resp = requests.get(url)
    if resp.status_code != 200:
        print(
            f'    Could not fetch data for {pkg_name} from PyPI (status {resp.status_code}).'
        )
        return (pkg_name, specified_min, None)
    pypi_data = resp.json()
    alternative = find_lowest_version_with_compatible_wheel(
        pkg_name, spec, pypi_data, required_py_ver, specified_min
    )
    if alternative:
        print(
            f'    ISSUE: {pkg_name}=={specified_min} does NOT provide a compatible wheel; consider bumping to at least version {alternative}'
        )
    else:
        print(
            f'    ISSUE: No version of {pkg_name} satisfying {spec} provides a compatible wheel for Python {required_py_ver}'
        )
    return (pkg_name, specified_min, alternative)


def process_all_dependencies(
    dependencies, required_py_ver: Version, origin: str
):
    """
    Processes a list of dependency strings (from either main dependencies or extras)
    and returns a list of issue reports.
    """
    issues = []
    for dep in dependencies:
        result = process_dependency(dep, required_py_ver, origin)
        if result:
            issues.append(result)
    return issues


def process_extras(extras: dict, required_py_ver: Version):
    """
    Processes extra dependencies (grouped by extra name) and returns a list of issues.
    """
    issues = []
    for extra_name, dep_list in extras.items():
        print(f'\n--- Processing extra: {extra_name} ---')
        issues.extend(
            process_all_dependencies(
                dep_list, required_py_ver, f"extra '{extra_name}'"
            )
        )
    return issues


# --- Main Script ---


def main(pyproject_path='pyproject.toml'):
    try:
        with open(pyproject_path, 'rb') as f:
            config = tomllib.load(f)
    except Exception as e:  # noqa: BLE001
        print(f'Error reading {pyproject_path}: {e}')
        sys.exit(1)

    project = config.get('project')
    if not project:
        print('No [project] section found in pyproject.toml.')
        sys.exit(1)

    requires_python = project.get('requires-python')
    if not requires_python:
        print('No requires-python field found in [project].')
        sys.exit(1)

    req_py_version = parse_min_python_version(requires_python)
    if req_py_version is None:
        print('Could not determine required Python version.')
        sys.exit(1)

    print(f'Required Python minimal version: {req_py_version}')

    overall_issues = []

    # Process main dependencies (list of strings)
    main_deps = project.get('dependencies', [])
    if main_deps:
        print('\n--- Checking main dependencies ---')
        overall_issues.extend(
            process_all_dependencies(main_deps, req_py_version, 'main')
        )
    else:
        print('No main dependencies found.')

    # Process extras (dictionary: extra name -> list of dependencies)
    extras = project.get('optional-dependencies', {})
    if extras:
        print('\n--- Checking extra dependencies ---')
        overall_issues.extend(process_extras(extras, req_py_version))
    else:
        print('No extra dependencies found.')

    if overall_issues:
        print(
            f'\nSummary of problematic binary dependencies (for required Python version {req_py_version}):'
        )
        for pkg_name, spec_min, alternative in overall_issues:
            if alternative:
                print(
                    f'  - {pkg_name}: {pkg_name}=={spec_min} does NOT provide a compatible wheel; consider bumping to at least {alternative}'
                )
            else:
                print(
                    f'  - {pkg_name}: {pkg_name}=={spec_min} does NOT provide a compatible wheel and no alternative was found'
                )
    else:
        print(
            '\nAll binary dependencies provide a compatible wheel for the required Python version based on their specified minimal version.'
        )


if __name__ == '__main__':
    main()



from qtpy.QtCore import QModelIndex, QRect
from qtpy.QtGui import QIcon, QImage, QPainter, QPixmap
from qtpy.QtWidgets import (
    QComboBox,
    QHBoxLayout,
    QLabel,
    QListView,
    QPushButton,
    QStyledItemDelegate,
    QStyleOptionViewItem,
    QWidget,
)

from napari._qt.layer_controls.widgets.qt_widget_controls_base import (
    QtWidgetControlsBase,
    QtWrappedLabel,
)
from napari.layers.base.base import Layer
from napari.utils.colormaps import (
    AVAILABLE_COLORMAPS,
    display_name_to_name,
    ensure_colormap,
    make_colorbar,
)
from napari.utils.translations import trans

COLORMAP_WIDTH = 50
TEXT_WIDTH = 130
ENTRY_HEIGHT = 20
PADDING = 1


class ColorStyledDelegate(QStyledItemDelegate):
    """Class for paint :py:class:`~.ColorComboBox` elements when list trigger

    Parameters
    ----------
    base_height : int
        Height of single list element.
    color_dict: dict
        Dict mapping name to colors.
    """

    def __init__(self, base_height: int, **kwargs) -> None:
        super().__init__(**kwargs)
        self.base_height = base_height

    def paint(
        self,
        painter: QPainter,
        style: QStyleOptionViewItem,
        model: QModelIndex,
    ):
        style2 = QStyleOptionViewItem(style)

        cbar_rect = QRect(
            style.rect.x(),
            style.rect.y() + PADDING,
            style.rect.width() - TEXT_WIDTH,
            style.rect.height() - 2 * PADDING,
        )
        text_rect = QRect(
            style.rect.width() - TEXT_WIDTH,
            style.rect.y() + PADDING,
            style.rect.width(),
            style.rect.height() - 2 * PADDING,
        )
        style2.rect = text_rect
        super().paint(painter, style2, model)
        name = display_name_to_name(model.data())
        cbar = make_colorbar(ensure_colormap(name), (18, 100))
        image = QImage(
            cbar,
            cbar.shape[1],
            cbar.shape[0],
            QImage.Format_RGBA8888,
        )
        painter.drawImage(cbar_rect, image)

    def sizeHint(self, style: QStyleOptionViewItem, model: QModelIndex):
        res = super().sizeHint(style, model)
        res.setHeight(self.base_height)
        res.setWidth(max(500, res.width()))
        return res


class QtColormapComboBox(QComboBox):
    """Combobox showing colormaps

    Parameters
    ----------
    parent : QWidget
        Parent widget of comboxbox.
    """

    def __init__(self, parent) -> None:
        super().__init__(parent)
        view = QListView()
        view.setMinimumWidth(COLORMAP_WIDTH + TEXT_WIDTH)
        view.setItemDelegate(ColorStyledDelegate(ENTRY_HEIGHT))
        self.setView(view)


class QtColormapControl(QtWidgetControlsBase):
    """
    Class that wraps the connection of events/signals between the layer colormaps
    attribute and Qt widgets.

    Parameters
    ----------
    parent: qtpy.QtWidgets.QWidget
        An instance of QWidget that will be used as widgets parent
    layer : napari.layers.Layer
        An instance of a napari layer.

    Attributes
    ----------
    colorbar_label : qtpy.QtWidgets.QPushButton
        Button showing colorbar widget. Also enables selection of custom colormap.
    colormap_combobox : qtpy.QtWidgets.QComboBox
        ComboBox controlling current colormap of the layer.
    colormapWidget : qtpy.QtWidgets.QWidget
        Widget to wrap combobox and label widgets related with the layer colormap attribute.
    colormap_widget_label : napari._qt.layer_controls.widgets.qt_widget_controls_base.QtWrappedLabel
        Label for the color mode chooser widget.
    """

    def __init__(self, parent: QWidget, layer: Layer) -> None:
        super().__init__(parent, layer)
        # Setup layer
        self._layer.events.colormap.connect(self._on_colormap_change)

        # Setup widgets
        comboBox = QtColormapComboBox(parent)
        comboBox.setObjectName('colormapComboBox')
        comboBox._allitems = set(self._layer.colormaps)

        for name, cm in AVAILABLE_COLORMAPS.items():
            if name in self._layer.colormaps:
                comboBox.addItem(cm._display_name, name)

        comboBox.currentTextChanged.connect(self.change_color)
        self.colormap_combobox = comboBox
        self.colorbar_label = QPushButton(parent=parent)
        self.colorbar_label.setObjectName('colorbar')
        self.colorbar_label.setToolTip(trans._('Colorbar'))
        self.colorbar_label.clicked.connect(self._on_make_colormap)

        colormap_layout = QHBoxLayout()
        colormap_layout.setContentsMargins(0, 0, 0, 2)
        if hasattr(self._layer, 'rgb') and self._layer.rgb:
            colormap_layout.addWidget(QLabel('RGB'))
            self.colormap_combobox.setVisible(False)
            self.colorbar_label.setVisible(False)
        else:
            colormap_layout.addWidget(self.colorbar_label)
            colormap_layout.addWidget(self.colormap_combobox, stretch=1)
        colormap_layout.addStretch(1)
        self.colormapWidget = QWidget()
        self.colormapWidget.setProperty('emphasized', True)
        self.colormapWidget.setLayout(colormap_layout)

        self._on_colormap_change()

        self.colormap_widget_label = QtWrappedLabel(trans._('colormap:'))

    def change_color(self, text):
        """Change colormap on the layer model.

        Parameters
        ----------
        text : str
            Colormap name.
        """
        self._layer.colormap = self.colormap_combobox.currentData()

    def _on_colormap_change(self):
        """Receive layer model colormap change event and update dropdown menu."""
        name = self._layer.colormap.name
        if name not in self.colormap_combobox._allitems and (
            cm := AVAILABLE_COLORMAPS.get(name)
        ):
            self.colormap_combobox._allitems.add(name)
            self.colormap_combobox.addItem(cm._display_name, name)

        if name != self.colormap_combobox.currentData():
            index = self.colormap_combobox.findData(name)
            self.colormap_combobox.setCurrentIndex(index)

        # Note that QImage expects the image width followed by height
        cbar = self._layer.colormap.colorbar
        image = QImage(
            cbar,
            cbar.shape[1],
            cbar.shape[0],
            QImage.Format_RGBA8888,
        )
        self.colorbar_label.setIcon(QIcon(QPixmap.fromImage(image)))

    def _on_make_colormap(self):
        """Make new colormap when colorbar_label (pushbutton) is pressed."""
        from napari._qt.utils import get_color
        from napari.utils.colormaps.colormap_utils import ensure_colormap

        color = get_color(self.parent(), mode='hex')
        if color:
            self._layer.colormap = ensure_colormap(color)

    def get_widget_controls(self) -> list[tuple[QtWrappedLabel, QWidget]]:
        return [(self.colormap_widget_label, self.colormapWidget)]



"""
nD labels
=========

Display a labels layer above of an image layer using the ``add_labels`` and
``add_image`` APIs

.. tags:: visualization-nD
"""

from scipy import ndimage as ndi
from skimage import data

import napari

blobs = data.binary_blobs(length=128, volume_fraction=0.1, n_dim=3)
viewer = napari.Viewer()
layer = viewer.add_image(blobs[::2].astype(float), name='blobs', scale=(2, 1, 1))
labeled = ndi.label(blobs)[0]
viewer.add_labels(labeled[::2], name='blob ID', scale=(2, 1, 1))

if __name__ == '__main__':
    napari.run()



# See "Writing benchmarks" in the asv docs for more information.
# https://asv.readthedocs.io/en/latest/writing_benchmarks.html
# or the napari documentation on benchmarking
# https://github.com/napari/napari/blob/main/docs/BENCHMARKS.md
from __future__ import annotations

import os

import numpy as np
from packaging.version import parse as parse_version
from qtpy.QtWidgets import QApplication

import napari
import napari.layers

NAPARI_0_4_19 = parse_version(napari.__version__) <= parse_version('0.4.19')


class QtViewerViewImageSuite:
    """Benchmarks for viewing images in the viewer."""

    data: np.ndarray
    viewer: napari.Viewer | None

    params = [2**i for i in range(4, 13)]

    if 'PR' in os.environ:
        skip_params = [(2**i,) for i in range(6, 13)]

    def setup(self, n):
        _ = QApplication.instance() or QApplication([])
        rng = np.random.default_rng(0)
        self.data = rng.random((n, n))
        self.viewer = None

    def teardown(self, n):
        self.viewer.window.close()

    def time_view_image(self, n):
        """Time to view an image."""
        self.viewer, _ = napari.imshow(self.data)


class QtViewerAddImageSuite:
    """Benchmarks for adding images to the viewer."""

    params = [2**i for i in range(4, 13)]

    if 'PR' in os.environ:
        skip_params = [(2**i,) for i in range(6, 13)]

    def setup(self, n):
        _ = QApplication.instance() or QApplication([])
        np.random.seed(0)
        self.data = np.random.random((n, n))
        self.viewer = napari.Viewer()

    def teardown(self, n):
        self.viewer.window.close()

    def time_add_image(self, n):
        """Time to view an image."""
        self.viewer.add_image(self.data)


class QtViewerImageSuite:
    """Benchmarks for images in the viewer."""

    viewer: napari.Viewer
    data: np.ndarray

    params = [2**i for i in range(4, 13)]

    if 'PR' in os.environ:
        skip_params = [(2**i,) for i in range(6, 13)]

    def setup(self, n):
        _ = QApplication.instance() or QApplication([])
        rng = np.random.default_rng(0)
        self.data = rng.random((n, n))
        self.viewer = napari.Viewer()
        self.viewer.add_image(self.data)

    def teardown(self, n):
        self.viewer.window.close()

    def time_zoom(self, n):
        """Time to zoom in and zoom out."""
        if NAPARI_0_4_19:
            self.viewer.window._qt_viewer.view.camera.zoom(
                0.5, center=(0.5, 0.5)
            )
            self.viewer.window._qt_viewer.view.camera.zoom(
                2.0, center=(0.5, 0.5)
            )
        else:
            self.viewer.window._qt_viewer.canvas.view.camera.zoom(
                0.5, center=(0.5, 0.5)
            )
            self.viewer.window._qt_viewer.canvas.view.camera.zoom(
                2.0, center=(0.5, 0.5)
            )

    def time_refresh(self, n):
        """Time to refresh view."""
        self.viewer.layers[0].refresh()

    def time_set_view_slice(self, n):
        """Time to set view slice."""
        layer = self.viewer.layers[0]
        if hasattr(layer, '_slicing_state'):
            layer._slicing_state._set_view_slice()
        else:
            # before https://github.com/napari/napari/pull/8254
            layer._set_view_slice()

    def time_update_thumbnail(self, n):
        """Time to update thumbnail."""
        self.viewer.layers[0]._update_thumbnail()

    def time_get_value(self, n):
        """Time to get current value."""
        self.viewer.layers[0].get_value((0,) * 2)


class QtViewerSingleImageSuite:
    """Benchmarks for a single image layer in the viewer."""

    data: np.ndarray
    new_data: np.ndarray
    viewer: napari.Viewer
    layer: napari.layers.Image

    def setup(self):
        _ = QApplication.instance() or QApplication([])
        rng = np.random.default_rng(0)
        self.data = rng.random((128, 128, 128))
        self.new_data = rng.random((128, 128, 128))
        self.viewer = napari.Viewer()
        self.layer = self.viewer.add_image(self.data)

    def teardown(self):
        self.viewer.window.close()

    def time_zoom(self):
        """Time to zoom in and zoom out."""
        if NAPARI_0_4_19:
            self.viewer.window._qt_viewer.view.camera.zoom(
                0.5, center=(0.5, 0.5)
            )
            self.viewer.window._qt_viewer.view.camera.zoom(
                2.0, center=(0.5, 0.5)
            )
        else:
            self.viewer.window._qt_viewer.canvas.view.camera.zoom(
                0.5, center=(0.5, 0.5)
            )
            self.viewer.window._qt_viewer.canvas.view.camera.zoom(
                2.0, center=(0.5, 0.5)
            )

    def time_set_data(self):
        """Time to set view slice."""
        self.layer.data = self.new_data

    def time_refresh(self):
        """Time to refresh view."""
        self.layer.refresh()

    def time_set_view_slice(self):
        """Time to set view slice."""
        if hasattr(self.layer, '_slicing_state'):
            self.layer._slicing_state._set_view_slice()
        else:
            # before https://github.com/napari/napari/pull/8254
            self.layer._set_view_slice()

    def time_update_thumbnail(self):
        """Time to update thumbnail."""
        self.layer._update_thumbnail()

    def time_get_value(self):
        """Time to get current value."""
        self.layer.get_value((0,) * 3)

    def time_ndisplay(self):
        """Time to enter 3D rendering."""
        self.viewer.dims.ndisplay = 3


class QtViewerSingleInvisbleImageSuite:
    """Benchmarks for a invisible single image layer in the viewer."""

    data: np.ndarray
    new_data: np.ndarray
    viewer: napari.Viewer
    layer: napari.layers.Image

    def setup(self):
        _ = QApplication.instance() or QApplication([])
        rng = np.random.default_rng(0)
        self.data = rng.random((128, 128, 128))
        self.new_data = rng.random((128, 128, 128))
        self.viewer = napari.Viewer()
        self.layer = self.viewer.add_image(self.data, visible=False)

    def teardown(self):
        self.viewer.window.close()

    def time_zoom(self):
        """Time to zoom in and zoom out."""
        if NAPARI_0_4_19:
            self.viewer.window._qt_viewer.view.camera.zoom(
                0.5, center=(0.5, 0.5)
            )
            self.viewer.window._qt_viewer.view.camera.zoom(
                2.0, center=(0.5, 0.5)
            )
        else:
            self.viewer.window._qt_viewer.canvas.view.camera.zoom(
                0.5, center=(0.5, 0.5)
            )
            self.viewer.window._qt_viewer.canvas.view.camera.zoom(
                2.0, center=(0.5, 0.5)
            )

    def time_set_data(self):
        """Time to set view slice."""
        self.layer.data = self.new_data

    def time_refresh(self):
        """Time to refresh view."""
        self.layer.refresh()

    def time_set_view_slice(self):
        """Time to set view slice."""
        if hasattr(self.layer, '_slicing_state'):
            self.layer._slicing_state._set_view_slice()
        else:
            # before https://github.com/napari/napari/pull/8254
            self.layer._set_view_slice()

    def time_update_thumbnail(self):
        """Time to update thumbnail."""
        self.layer._update_thumbnail()

    def time_get_value(self):
        """Time to get current value."""
        self.layer.get_value((0,) * 3)

    def time_ndisplay(self):
        """Time to enter 3D rendering."""
        self.viewer.dims.ndisplay = 3


class QtImageRenderingSuite:
    """Benchmarks for a single image layer in the viewer."""

    data: np.ndarray
    viewer: napari.Viewer
    layer: napari.layers.Image

    params = [2**i for i in range(4, 13)]

    if 'PR' in os.environ:
        skip_params = [(2**i,) for i in range(6, 13)]

    def setup(self, n):
        _ = QApplication.instance() or QApplication([])
        rng = np.random.default_rng(0)
        self.data = rng.random((n, n)) * 2**12
        self.viewer = napari.Viewer(ndisplay=2)
        self.layer = self.viewer.add_image(self.data)

    def teardown(self, n):
        self.viewer.close()

    def time_change_contrast(self, n):
        """Time to change contrast limits."""
        self.layer.contrast_limits = (250, 3000)
        self.layer.contrast_limits = (300, 2900)
        self.layer.contrast_limits = (350, 2800)

    def time_change_gamma(self, n):
        """Time to change gamma."""
        self.layer.gamma = 0.5
        self.layer.gamma = 0.8
        self.layer.gamma = 1.3


class QtVolumeRenderingSuite:
    """Benchmarks for a single image layer in the viewer."""

    data: np.ndarray
    viewer: napari.Viewer

    params = [2**i for i in range(4, 10)]

    if 'PR' in os.environ:
        skip_params = [(2**i,) for i in range(6, 10)]

    def setup(self, n):
        _ = QApplication.instance() or QApplication([])
        rng = np.random.default_rng(0)
        self.data = rng.random((n, n, n)) * 2**12
        self.viewer = napari.Viewer(ndisplay=3)
        self.viewer.add_image(self.data)

    def teardown(self, n):
        self.viewer.close()

    def time_change_contrast(self, n):
        """Time to change contrast limits."""
        self.viewer.layers[0].contrast_limits = (250, 3000)
        self.viewer.layers[0].contrast_limits = (300, 2900)
        self.viewer.layers[0].contrast_limits = (350, 2800)

    def time_change_gamma(self, n):
        """Time to change gamma."""
        self.viewer.layers[0].gamma = 0.5
        self.viewer.layers[0].gamma = 0.8
        self.viewer.layers[0].gamma = 1.3


if __name__ == '__main__':
    from utils import run_benchmark

    run_benchmark()



"""
Visualizing optical flow in napari
==================================

Adapted from the scikit-image gallery [1]_.

In napari, we can show the flowing vortex as an additional dimension in the
image, visible by moving the slider.

.. tags:: visualization-advanced, layers

.. [1] https://scikit-image.org/docs/stable/auto_examples/registration/plot_opticalflow.html
"""
import numpy as np
from skimage.data import vortex
from skimage.registration import optical_flow_ilk

import napari

#######################################################################
# First, we load the vortex image as a 3D array. (time, row, column)

vortex_im = np.asarray(vortex())

#######################################################################
# We compute the optical flow using scikit-image. (Note: as of
# scikit-image 0.21, there seems to be a transposition of the image in
# the output, which we account for later.)

u, v = optical_flow_ilk(vortex_im[0], vortex_im[1], radius=15)

#######################################################################
# Compute the flow magnitude, for visualization.

magnitude = np.sqrt(u ** 2 + v ** 2)

#######################################################################
# We subsample the vector field to display it — it's too
# messy otherwise! And we transpose the rows/columns axes to match the
# current scikit-image output.

nvec = 21
nr, nc = magnitude.shape
step = max(nr//nvec, nc//nvec)
offset = step // 2
usub = u[offset::step, offset::step]
vsub = v[offset::step, offset::step]

vectors_field = np.stack([usub, vsub], axis=-1)

#######################################################################
# Finally, we create a viewer, and add the vortex frames, the flow
# magnitude, and the vector field.

viewer, vortex_layer = napari.imshow(vortex_im)
mag_layer = viewer.add_image(magnitude, colormap='magma', opacity=0.3)
flow_layer = viewer.add_vectors(
        vectors_field,
        name='optical flow',
        scale=[step, step],
        translate=[offset, offset],
        edge_width=0.3,
        length=0.3,
        )

if __name__ == '__main__':
    napari.run()



"""
nD surface
==========

Display a 3D surface

.. tags:: visualization-nD
"""

import numpy as np

import napari

# create the viewer and window
viewer = napari.Viewer(ndisplay=3)

data = np.array([[0, 0, 0], [0, 20, 10], [10, 0, -10], [10, 10, -10]])
faces = np.array([[0, 1, 2], [1, 2, 3]])
values = np.linspace(0, 1, len(data))

# add the surface
layer = viewer.add_surface((data, faces, values))

if __name__ == '__main__':
    napari.run()



"""
Multithreading simple
=====================

.. tags:: interactivity
"""

import time

from qtpy.QtWidgets import QApplication, QHBoxLayout, QLabel, QWidget

from napari.qt import thread_worker


@thread_worker
def long_running_function():
    """Just a long running function, most like viewer.update."""
    time.sleep(2)  # long function
    return 'finished!'


def create_widget():
    widget = QWidget()
    layout = QHBoxLayout()
    widget.setLayout(layout)
    widget.status = QLabel('ready...')
    layout.addWidget(widget.status)
    widget.show()
    return widget


if __name__ == '__main__':
    app = QApplication([])
    wdg = create_widget()

    # call decorated function
    # By default, @thread_worker-decorated functions do not immediately start
    worker = long_running_function()
    # Signals are best connected *before* starting the worker.
    worker.started.connect(lambda: wdg.status.setText('worker is running...'))
    worker.returned.connect(lambda x: wdg.status.setText(f'returned {x}'))

    # # Connections may also be passed directly to the decorated function.
    # # The above syntax is equivalent to:
    # worker = long_running_function(
    #     _connect={
    #         'started': lambda: wdg.status.setText("worker is running..."),
    #         'returned': lambda x: wdg.status.setText(f"returned {x!r}"),
    #     }
    # )

    worker.start()

    app.exec_()



import sys

names = ['MAIN', 'SECOND', 'THIRD', 'FOURTH']
num = int(sys.argv[1])
values = sys.argv[2].split(',')

if num < len(values):
    print(f'{names[num]}={values[num]}')
else:
    print(f'{names[num]}=none')



"""
Multiple viewer widget
======================

This is an example of how to have more than one viewer in the same napari window.
Additional viewers state will be synchronized with the main viewer.
Switching to 3D display will only impact the main viewer.

This example also contains the option to enable cross that will be moved to the
current dims point (`viewer.dims.point`).

.. tags:: gui
"""

from copy import deepcopy

import numpy as np
from qtpy.QtCore import Qt
from qtpy.QtWidgets import (
    QCheckBox,
    QDoubleSpinBox,
    QPushButton,
    QSplitter,
    QTabWidget,
    QVBoxLayout,
    QWidget,
)
from superqt.utils import qthrottled

import napari
from napari.components.viewer_model import ViewerModel
from napari.layers import Labels, Layer, Vectors
from napari.qt import QtViewer
from napari.utils.action_manager import action_manager
from napari.utils.events.event import WarningEmitter
from napari.utils.notifications import show_info


def copy_layer(layer: Layer, name: str = ''):
    res_layer = Layer.create(*layer.as_layer_data_tuple())
    res_layer.metadata['viewer_name'] = name
    return res_layer


def get_property_names(layer: Layer):
    klass = layer.__class__
    res = []
    for event_name, event_emitter in layer.events.emitters.items():
        if isinstance(event_emitter, WarningEmitter):
            continue
        if event_name in ('thumbnail', 'name'):
            continue
        if (
            isinstance(getattr(klass, event_name, None), property)
            and getattr(klass, event_name).fset is not None
        ):
            res.append(event_name)
    return res


def center_cross_on_mouse(
    viewer_model: napari.components.viewer_model.ViewerModel,
):
    """move the cross to the mouse position"""

    if not getattr(viewer_model, 'mouse_over_canvas', True):
        # There is no way for napari 0.4.15 to check if mouse is over sending canvas.
        show_info(
            'Mouse is not over the canvas. You may need to click on the canvas.'
        )
        return

    viewer_model.dims.current_step = tuple(
        np.round(
            [
                max(min_, min(p, max_)) / step
                for p, (min_, max_, step) in zip(
                    viewer_model.cursor.position, viewer_model.dims.range, strict=False
                )
            ]
        ).astype(int)
    )


action_manager.register_action(
    name='napari:move_point',
    command=center_cross_on_mouse,
    description='Move dims point to mouse position',
    keymapprovider=ViewerModel,
)

action_manager.bind_shortcut('napari:move_point', 'C')


class own_partial:
    """
    Workaround for deepcopy not copying partial functions
    (Qt widgets are not serializable)
    """

    def __init__(self, func, *args, **kwargs) -> None:
        self.func = func
        self.args = args
        self.kwargs = kwargs

    def __call__(self, *args, **kwargs):
        return self.func(*(self.args + args), **{**self.kwargs, **kwargs})

    def __deepcopy__(self, memodict=None):
        if memodict is None:
            memodict = {}
        return own_partial(
            self.func,
            *deepcopy(self.args, memodict),
            **deepcopy(self.kwargs, memodict),
        )


class QtViewerWrap(QtViewer):
    def __init__(self, main_viewer, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.main_viewer = main_viewer

    def _qt_open(
        self,
        filenames: list,
        stack: bool,
        plugin: str | None = None,
        layer_type: str | None = None,
        **kwargs,
    ):
        """for drag and drop open files"""
        self.main_viewer.window._qt_viewer._qt_open(
            filenames, stack, plugin, layer_type, **kwargs
        )


class CrossWidget(QCheckBox):
    """
    Widget to control the layer representing cross.
    Because of the performance reasons,
    the update of cross is throttled
    """

    def __init__(self, viewer: napari.Viewer) -> None:
        super().__init__('Add cross layer')
        self.viewer = viewer
        self.setChecked(False)
        self.stateChanged.connect(self._update_cross_visibility)
        self.layer = None
        self.viewer.dims.events.order.connect(self.update_cross)
        self.viewer.dims.events.ndim.connect(self._update_ndim)
        self.viewer.dims.events.current_step.connect(self.update_cross)
        self._extent = None

        self._update_extent()
        self.viewer.dims.events.connect(self._update_extent)

    @qthrottled(leading=False)
    def _update_extent(self):
        """
        Calculate the extent of the data.

        Ignores the layer with cross itself in calculating the extent.
        """
        layers = [
            layer
            for layer in self.viewer.layers
            if layer is not self.layer
        ]
        self._extent = self.viewer.layers.get_extent(layers)
        self.update_cross()

    def _update_ndim(self, event):
        if self.layer in self.viewer.layers:
            self.viewer.layers.remove(self.layer)
        self.layer = Vectors(name='.cross', ndim=event.value)
        self.layer.edge_width = 1.5
        self.update_cross()

    def _update_cross_visibility(self, state):
        if state:
            self.viewer.layers.append(self.layer)
        else:
            self.viewer.layers.remove(self.layer)
        self.update_cross()

    def update_cross(self):
        if self.layer not in self.viewer.layers:
            return

        point = self.viewer.dims.current_step
        vec = []
        for i, (lower, upper) in enumerate(self._extent.world.T):
            if (upper - lower) / self._extent.step[i] == 1:
                continue
            point1 = list(point)
            point1[i] = (lower + self._extent.step[i] / 2) / self._extent.step[
                i
            ]
            point2 = [0 for _ in point]
            point2[i] = (upper - lower) / self._extent.step[i]
            vec.append((point1, point2))
        if np.any(self.layer.scale != self._extent.step):
            self.layer.scale = self._extent.step
        self.layer.data = vec


class ExampleWidget(QWidget):
    """
    Example widget showcasing how to place additional widgets to the right
    of the additional viewers.
    """

    def __init__(self) -> None:
        super().__init__()
        self.btn = QPushButton('Perform action')
        self.spin = QDoubleSpinBox()
        layout = QVBoxLayout()
        layout.addWidget(self.spin)
        layout.addWidget(self.btn)
        layout.addStretch(1)
        self.setLayout(layout)


class MultipleViewerWidget(QSplitter):
    """The main widget of the example."""

    def __init__(self, viewer: napari.Viewer) -> None:
        super().__init__()
        self.viewer = viewer
        self.viewer_model1 = ViewerModel(title='model1')
        self.viewer_model2 = ViewerModel(title='model2')
        self._block = False
        self.qt_viewer1 = QtViewerWrap(viewer, self.viewer_model1)
        self.qt_viewer2 = QtViewerWrap(viewer, self.viewer_model2)
        self.tab_widget = QTabWidget()
        w1 = ExampleWidget()
        w2 = ExampleWidget()
        self.tab_widget.addTab(w1, 'Sample 1')
        self.tab_widget.addTab(w2, 'Sample 2')
        viewer_splitter = QSplitter()
        viewer_splitter.setOrientation(Qt.Orientation.Vertical)
        viewer_splitter.addWidget(self.qt_viewer1)
        viewer_splitter.addWidget(self.qt_viewer2)
        viewer_splitter.setContentsMargins(0, 0, 0, 0)

        self.addWidget(viewer_splitter)
        self.addWidget(self.tab_widget)

        self.viewer.layers.events.inserted.connect(self._layer_added)
        self.viewer.layers.events.removed.connect(self._layer_removed)
        self.viewer.layers.events.moved.connect(self._layer_moved)
        self.viewer.layers.selection.events.active.connect(
            self._layer_selection_changed
        )
        self.viewer.dims.events.current_step.connect(self._point_update)
        self.viewer_model1.dims.events.current_step.connect(self._point_update)
        self.viewer_model2.dims.events.current_step.connect(self._point_update)
        self.viewer.dims.events.order.connect(self._order_update)
        self.viewer.events.reset_view.connect(self._reset_view)
        self.viewer_model1.events.status.connect(self._status_update)
        self.viewer_model2.events.status.connect(self._status_update)

    def _status_update(self, event):
        self.viewer.status = event.value

    def _reset_view(self):
        self.viewer_model1.reset_view()
        self.viewer_model2.reset_view()

    def _layer_selection_changed(self, event):
        """
        update of current active layer
        """
        if self._block:
            return

        if event.value is None:
            self.viewer_model1.layers.selection.active = None
            self.viewer_model2.layers.selection.active = None
            return

        self.viewer_model1.layers.selection.active = self.viewer_model1.layers[
            event.value.name
        ]
        self.viewer_model2.layers.selection.active = self.viewer_model2.layers[
            event.value.name
        ]

    def _point_update(self, event):
        for model in [self.viewer, self.viewer_model1, self.viewer_model2]:
            if model.dims is event.source:
                continue
            if len(self.viewer.layers) != len(model.layers):
                continue
            model.dims.current_step = event.value

    def _order_update(self):
        order = list(self.viewer.dims.order)
        if len(order) <= 2:
            self.viewer_model1.dims.order = order
            self.viewer_model2.dims.order = order
            return

        order[-3:] = order[-2], order[-3], order[-1]
        self.viewer_model1.dims.order = tuple(order)
        order = list(self.viewer.dims.order)
        order[-3:] = order[-1], order[-2], order[-3]
        self.viewer_model2.dims.order = tuple(order)

    def _layer_added(self, event):
        """add layer to additional viewers and connect all required events"""
        self.viewer_model1.layers.insert(
            event.index, copy_layer(event.value, 'model1')
        )
        self.viewer_model2.layers.insert(
            event.index, copy_layer(event.value, 'model2')
        )
        for name in get_property_names(event.value):
            getattr(event.value.events, name).connect(
                own_partial(self._property_sync, name)
            )

        if isinstance(event.value, Labels):
            event.value.events.set_data.connect(self._set_data_refresh)
            event.value.events.labels_update.connect(self._set_data_refresh)
            self.viewer_model1.layers[
                event.value.name
            ].events.set_data.connect(self._set_data_refresh)
            self.viewer_model2.layers[
                event.value.name
            ].events.set_data.connect(self._set_data_refresh)
            event.value.events.labels_update.connect(self._set_data_refresh)
            self.viewer_model1.layers[
                event.value.name
            ].events.labels_update.connect(self._set_data_refresh)
            self.viewer_model2.layers[
                event.value.name
            ].events.labels_update.connect(self._set_data_refresh)
        if event.value.name != '.cross':
            self.viewer_model1.layers[event.value.name].events.data.connect(
                self._sync_data
            )
            self.viewer_model2.layers[event.value.name].events.data.connect(
                self._sync_data
            )

        event.value.events.name.connect(self._sync_name)

        self._order_update()

    def _sync_name(self, event):
        """sync name of layers"""
        index = self.viewer.layers.index(event.source)
        self.viewer_model1.layers[index].name = event.source.name
        self.viewer_model2.layers[index].name = event.source.name

    def _sync_data(self, event):
        """sync data modification from additional viewers"""
        if self._block:
            return
        for model in [self.viewer, self.viewer_model1, self.viewer_model2]:
            layer = model.layers[event.source.name]
            if layer is event.source:
                continue
            try:
                self._block = True
                layer.data = event.source.data
            finally:
                self._block = False

    def _set_data_refresh(self, event):
        """
        synchronize data refresh between layers
        """
        if self._block:
            return
        for model in [self.viewer, self.viewer_model1, self.viewer_model2]:
            layer = model.layers[event.source.name]
            if layer is event.source:
                continue
            try:
                self._block = True
                layer.refresh()
            finally:
                self._block = False

    def _layer_removed(self, event):
        """remove layer in all viewers"""
        self.viewer_model1.layers.pop(event.index)
        self.viewer_model2.layers.pop(event.index)

    def _layer_moved(self, event):
        """update order of layers"""
        dest_index = (
            event.new_index
            if event.new_index < event.index
            else event.new_index + 1
        )
        self.viewer_model1.layers.move(event.index, dest_index)
        self.viewer_model2.layers.move(event.index, dest_index)

    def _property_sync(self, name, event):
        """Sync layers properties (except the name)"""
        if event.source not in self.viewer.layers:
            return
        try:
            self._block = True
            setattr(
                self.viewer_model1.layers[event.source.name],
                name,
                getattr(event.source, name),
            )
            setattr(
                self.viewer_model2.layers[event.source.name],
                name,
                getattr(event.source, name),
            )
        finally:
            self._block = False


if __name__ == '__main__':
    from qtpy import QtWidgets
    QtWidgets.QApplication.setAttribute(Qt.ApplicationAttribute.AA_ShareOpenGLContexts)
    # above two lines are needed to allow undocking the widget with
    # additional viewers
    view = napari.Viewer()
    dock_widget = MultipleViewerWidget(view)
    cross = CrossWidget(view)

    view.window.add_dock_widget(dock_widget, name='Sample')
    view.window.add_dock_widget(cross, name='Cross', area='left')

    view.open_sample('napari', 'cells3d')

    napari.run()



"""
Affine transforms
=================

Display an image and its corners before and after an affine transform

.. tags:: visualization-advanced
"""
import numpy as np
import scipy.ndimage as ndi

import napari

# Create a random image
image = np.random.random((5, 5))

# Define an affine transform
affine = np.array([[1, -1, 4], [2, 3, 2], [0, 0, 1]])

# Define the corners of the image, including in homogeneous space
corners = np.array([[0, 0], [4, 0], [0, 4], [4, 4]])
corners_h = np.concatenate([corners, np.ones((4, 1))], axis=1)

viewer = napari.Viewer()

# Add the original image and its corners
viewer.add_image(image, name='background', colormap='red', opacity=.5)
viewer.add_points(corners_h[:, :-1], size=0.5, opacity=.5, face_color=[0.8, 0, 0, 0.8], name='bg corners')

# Add another copy of the image, now with a transform, and add its transformed corners
viewer.add_image(image, colormap='blue', opacity=.5, name='moving', affine=affine)
viewer.add_points((corners_h @ affine.T)[:, :-1], size=0.5, opacity=.5, face_color=[0, 0, 0.8, 0.8], name='mv corners')

# Note how the transformed corner points remain at the corners of the transformed image

# Now add the a regridded version of the image transformed with scipy.ndimage.affine_transform
# Note that we have to use the inverse of the affine as scipy does `pull` (or `backward`) resampling,
# transforming the output space to the input to locate data, but napari does `push` (or `forward`) direction,
# transforming input to output.
scipy_affine = ndi.affine_transform(image, np.linalg.inv(affine), output_shape=(10, 25), order=5)
viewer.add_image(scipy_affine, colormap='green', opacity=.5, name='scipy')

# Reset the view
viewer.reset_view()

if __name__ == '__main__':
    napari.run()



from qtpy.QtCore import Qt
from qtpy.QtWidgets import QComboBox, QWidget
from superqt import QLabeledDoubleSlider

from napari._qt.layer_controls.widgets.qt_widget_controls_base import (
    QtWidgetControlsBase,
    QtWrappedLabel,
)
from napari._qt.utils import attr_to_settr, qt_signals_blocked
from napari.layers import Image
from napari.layers.image._image_constants import (
    ImageRendering,
)
from napari.utils.events.event_utils import connect_setattr
from napari.utils.translations import trans


class QtImageRenderControl(QtWidgetControlsBase):
    """
    Class that wraps the connection of events/signals between the layer attribute for
    the method to render an image and Qt widgets.

    Parameters
    ----------
    parent: qtpy.QtWidgets.QWidget
        An instance of QWidget that will be used as widgets parent
    layer : napari.layers.Image
        An instance of a napari Image layer.

    Attributes
    ----------
    render_combobox : qtpy.QtWidgets.QComboBox
        Combobox to control labels render method.
    render_label : napari._qt.layer_controls.widgets.qt_widget_controls_base.QtWrappedLabel
        Label for the way labels should be rendered chooser widget.
    iso_threshold_slider : superqt.QLabeledDoubleSlider
        Slider controlling the isosurface threshold value for rendering.
    iso_threshold_label : qtpy.QtWidgets.QLabel
        Label for the isosurface threshold slider widget.
    attenuation_slider : superqt.QLabeledDoubleSlider
        Slider controlling attenuation rate for `attenuated_mip` mode.
    attenuation_label : qtpy.QtWidgets.QLabel
        Label for the attenuation slider widget.
    """

    def __init__(self, parent: QWidget, layer: Image) -> None:
        super().__init__(parent, layer)
        # Setup layer
        self._layer.events.rendering.connect(self._on_rendering_change)
        self._layer.events.contrast_limits.connect(
            self._on_contrast_limits_change
        )

        # Setup widgets
        render_combobox = QComboBox(parent)
        rendering_options = [i.value for i in ImageRendering]
        render_combobox.addItems(rendering_options)
        index = render_combobox.findText(
            self._layer.rendering, Qt.MatchFlag.MatchFixedString
        )
        render_combobox.setCurrentIndex(index)
        render_combobox.currentTextChanged.connect(self.change_rendering)
        self.render_combobox = render_combobox

        self.render_label = QtWrappedLabel(trans._('rendering:'))

        sld = QLabeledDoubleSlider(Qt.Orientation.Horizontal, parent=parent)
        sld.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        cmin, cmax = self._layer.contrast_limits_range
        sld.setMinimum(cmin)
        sld.setMaximum(cmax)
        sld.setValue(self._layer.iso_threshold)
        connect_setattr(sld.valueChanged, self._layer, 'iso_threshold')
        self._callbacks.append(
            attr_to_settr(self._layer, 'iso_threshold', sld, 'setValue')
        )
        self.iso_threshold_slider = sld

        self.iso_threshold_label = QtWrappedLabel(trans._('iso threshold:'))

        sld = QLabeledDoubleSlider(Qt.Orientation.Horizontal, parent=parent)
        sld.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        sld.setMinimum(0)
        sld.setMaximum(0.5)
        sld.setSingleStep(0.001)
        sld.setValue(self._layer.attenuation)
        sld.setDecimals(3)
        connect_setattr(sld.valueChanged, self._layer, 'attenuation')
        self._callbacks.append(
            attr_to_settr(self._layer, 'attenuation', sld, 'setValue')
        )
        self.attenuation_slider = sld

        self.attenuation_label = QtWrappedLabel(trans._('attenuation:'))

    def change_rendering(self, text):
        """Change rendering mode for image display.

        Parameters
        ----------
        text : str
            Rendering mode used by vispy.
            Selects a preset rendering mode in vispy that determines how
            volume is displayed:
            * translucent: voxel colors are blended along the view ray until
              the result is opaque.
            * mip: maximum intensity projection. Cast a ray and display the
              maximum value that was encountered.
            * additive: voxel colors are added along the view ray until
              the result is saturated.
            * iso: isosurface. Cast a ray until a certain threshold is
              encountered. At that location, lighning calculations are
              performed to give the visual appearance of a surface.
            * attenuated_mip: attenuated maximum intensity projection. Cast a
              ray and attenuate values based on integral of encountered values,
              display the maximum value that was encountered after attenuation.
              This will make nearer objects appear more prominent.
        """
        self._layer.rendering = text
        self._update_rendering_parameter_visibility()

    def _on_rendering_change(self):
        """Receive layer model rendering change event and update dropdown menu."""
        with qt_signals_blocked(self.render_combobox):
            index = self.render_combobox.findText(
                self._layer.rendering, Qt.MatchFlag.MatchFixedString
            )
            self.render_combobox.setCurrentIndex(index)
            self._update_rendering_parameter_visibility()

    def _on_contrast_limits_change(self):
        with qt_signals_blocked(self.iso_threshold_slider):
            cmin, cmax = self._layer.contrast_limits_range
            self.iso_threshold_slider.setMinimum(cmin)
            self.iso_threshold_slider.setMaximum(cmax)

    def _on_display_change_hide(self):
        self.iso_threshold_slider.hide()
        self.iso_threshold_label.hide()
        self.attenuation_slider.hide()
        self.attenuation_label.hide()
        self.render_combobox.hide()
        self.render_label.hide()

    def _on_display_change_show(self):
        self.render_combobox.show()
        self.render_label.show()
        self._update_rendering_parameter_visibility()

    def _update_rendering_parameter_visibility(self):
        """Hide isosurface rendering parameters if they aren't needed."""
        rendering = ImageRendering(self._layer.rendering)
        iso_threshold_visible = rendering == ImageRendering.ISO
        self.iso_threshold_label.setVisible(iso_threshold_visible)
        self.iso_threshold_slider.setVisible(iso_threshold_visible)
        attenuation_visible = rendering == ImageRendering.ATTENUATED_MIP
        self.attenuation_slider.setVisible(attenuation_visible)
        self.attenuation_label.setVisible(attenuation_visible)

    def get_widget_controls(self) -> list[tuple[QtWrappedLabel, QWidget]]:
        return [
            (self.render_label, self.render_combobox),
            (self.iso_threshold_label, self.iso_threshold_slider),
            (self.attenuation_label, self.attenuation_slider),
        ]



"""
Point cloud
===========

Display 3D points with combinations of different renderings.

.. tags:: visualization-basic
"""

import numpy as np

import napari

n_points = 100

points = np.random.normal(10, 100, (n_points, 3))
symbols = np.random.choice(['o', 's', '*'], n_points)
sizes = np.random.rand(n_points) * 10 + 10
colors = np.random.rand(n_points, 3)

viewer = napari.Viewer(ndisplay=3)
viewer.add_points(points, symbol=symbols, size=sizes, face_color=colors)

if __name__ == '__main__':
    napari.run()



"""
Custom image interpolation kernels
==================================

When interpolation is set to 'custom', the convolution kernel provided by
`custom_interpolation_kernel_2d` is used to convolve the image on the gpu.
In this example, we use custom gaussian kernels of arbitrary size, a sharpening
kernel and a ridge detection kernel.

Under the hood, this works by by sampling the image texture with `linear`
interpolation in a regular grid (of size = of the kernel) around each fragment,
and then using the weights in the kernel to add up the final fragment value.

.. tags:: gui, visualization-nD

"""

import numpy as np
from magicgui import magicgui
from scipy.signal.windows import gaussian
from skimage import data

import napari

viewer = napari.Viewer()
layer = viewer.add_image(data.astronaut(), rgb=True, interpolation2d='custom')


def gaussian_kernel(size, sigma):
    window = gaussian(size, sigma)
    kernel = np.outer(window, window)
    return kernel / kernel.sum()


def sharpen_kernel():
    return np.array([
        [ 0, -1,  0],
        [-1,  5, -1],
        [ 0, -1,  0],
    ])


def ridge_detection_kernel():
    return np.array([
        [-1, -1, -1],
        [-1,  9, -1],
        [-1, -1, -1],
    ])


@magicgui(
    auto_call=True,
    kernel_size={'widget_type': 'Slider', 'min': 1, 'max': 20},
    sigma={'widget_type': 'FloatSlider', 'min': 0.1, 'max': 5, 'step': 0.1},
    kernel_type={'choices': ['none', 'gaussian', 'sharpen', 'ridge_detection']},
)
def gpu_kernel(image: napari.layers.Image, kernel_type: str = 'gaussian', kernel_size: int = 5, sigma: float = 1):
    if kernel_type == 'none':
        image.interpolation2d = 'linear'
    else:
        image.interpolation2d = 'custom'

    if kernel_type == 'gaussian':
        gpu_kernel.kernel_size.show()
        gpu_kernel.sigma.show()
    else:
        gpu_kernel.kernel_size.hide()
        gpu_kernel.sigma.hide()

    if kernel_type == 'gaussian':
        image.custom_interpolation_kernel_2d = gaussian_kernel(kernel_size, sigma)
    elif kernel_type == 'sharpen':
        image.custom_interpolation_kernel_2d = sharpen_kernel()
    elif kernel_type == 'ridge_detection':
        image.custom_interpolation_kernel_2d = ridge_detection_kernel()


viewer.window.add_dock_widget(gpu_kernel)
gpu_kernel()


if __name__ == '__main__':
    napari.run()



"""
Add points
==========

Display a points layer on top of an image layer using the ``add_points`` and
``add_image`` APIs

.. tags:: visualization-basic
"""

import numpy as np
from skimage import data
from skimage.color import rgb2gray

import napari

# add the image
viewer = napari.Viewer()
layer = viewer.add_image(rgb2gray(data.astronaut()))
# add the points
points = np.array([[100, 100], [200, 200], [333, 111]])
size = np.array([10, 20, 20])
viewer.add_points(points, size=size)

# unselect the image layer
viewer.layers.selection.discard(viewer.layers[0])

# adjust some of the points layer attributes
layer = viewer.layers[1]

# change the layer name
layer.name = 'points'

# change the layer visibility
layer.visible = False
layer.visible = True

# select the layer
viewer.layers.selection.add(layer)
# deselect the layer
viewer.layers.selection.remove(layer)
# or: viewer.layers.selection.discard(layer)

# change the layer opacity
layer.opacity = 0.9

# change the layer point symbol using an alias
layer.symbol = '+'

# change the layer point out_of_slice_display status
layer.out_of_slice_display = True

# change the layer mode
layer.mode = 'add'

if __name__ == '__main__':
    napari.run()



"""
Spheres
=======

Display two spheres with Surface layers

.. tags:: visualization-advanced
"""

from vispy.geometry import create_sphere

import napari

mesh = create_sphere(method='ico')

faces = mesh.get_faces()
vert = mesh.get_vertices() * 100

sphere1 = (vert + 30, faces)
sphere2 = (vert - 30, faces)

viewer = napari.Viewer(ndisplay=3)
surface1 = viewer.add_surface(sphere1)
surface2 = viewer.add_surface(sphere2)
viewer.reset_view()

if __name__ == '__main__':
    napari.run()



"""
Progress bar threading
======================

This file provides a minimal working example using a progress bar alongside
``@thread_worker`` to report progress.

.. tags:: interactivity
"""
from time import sleep

from qtpy.QtWidgets import QPushButton, QVBoxLayout, QWidget

import napari
from napari.qt import thread_worker

viewer = napari.Viewer()


def handle_yields(yielded_val):
    print(f'Just yielded: {yielded_val}')


# generator thread workers can provide progress updates on each yield
@thread_worker(
    # passing a progress dictionary with the total number of expected yields
    # will place a progress bar in the activity dock and increment its value
    # with each yield. We can optionally pass a description for the bar
    # using the 'desc' key.
    progress={'total': 5, 'desc': 'thread-progress'},
    # this does not preclude us from connecting other functions to any of the
    # worker signals (including `yielded`)
    connect={'yielded': handle_yields},
)
def my_long_running_thread(*_):
    for i in range(5):
        sleep(0.1)
        yield i


@thread_worker(
    # If we are unsure of the number of expected yields,
    # we can still pass an estimate to total,
    # and the progress bar will become indeterminate
    # once this number is exceeded.
    progress={'total': 5},
    # we can also get a simple indeterminate progress bar
    # by passing progress=True
    connect={'yielded': handle_yields},
)
def my_indeterminate_thread(*_):
    for i in range(10):
        sleep(0.1)
        yield i


def return_func(return_val):
    print(f'Returned: {return_val}')


# finally, a FunctionWorker can still provide an indeterminate
# progress bar, but will not take a total>0
@thread_worker(
    progress={'total': 0, 'desc': 'FunctionWorker'},
    # can use progress=True if not passing description
    connect={'returned': return_func},
)
def my_function(*_):
    sum_val = 0
    for i in range(10):
        sum_val += i
        sleep(0.1)
    return sum_val


button_layout = QVBoxLayout()
start_btn = QPushButton('Start')
start_btn.clicked.connect(my_long_running_thread)
button_layout.addWidget(start_btn)

start_btn2 = QPushButton('Start Indeterminate')
start_btn2.clicked.connect(my_indeterminate_thread)
button_layout.addWidget(start_btn2)

start_btn3 = QPushButton('Start FunctionWorker')
start_btn3.clicked.connect(my_function)
button_layout.addWidget(start_btn3)

pbar_widget = QWidget()
pbar_widget.setLayout(button_layout)
pbar_widget.setObjectName('Threading Examples')
viewer.window.add_dock_widget(pbar_widget, allowed_areas=['right'])

# showing the activity dock so we can see the progress bars
viewer.window._status_bar._toggle_activity_dock(True)

if __name__ == '__main__':
    napari.run()



"""
Surface with multiple textures
==============================

This example demonstrates one possible method for displaying a 3D surface with
multiple textures.

Thanks to Emmanuel Reynaud and Luis Gutierrez for providing the gorgeous coral
model for this demo. You can find the data on FigShare:
https://zenodo.org/records/13380203

More information on the methods used to generate this model can be found in *L.
Gutierrez-Heredia, C. Keogh, E. G. Reynaud, Assessing the Capabilities of
Additive Manufacturing Technologies for Coral Studies, Education, and
Monitoring. Front. Mar. Sci. 5 (2018), doi:10.3389/fmars.2018.00278.*

A bit about 3D models
---------------------

A standard way to define a 3D model (mesh, or Surface in napari) is by listing
vertices (3D point coordinates) and faces (triplets of vertex indices - each
face is a triangle in 3D space). Meshes are often stored in "Wavefront" (.obj)
files, which may have companion material (.mtl) files that describe some
shading properties (base color, shinyness, etc.) for different parts of the
model.

In some cases, the color of a vertex is given by a single point value that is
then colormapped on the fly (`vertex_values`). In other cases, each vertex or
face may be assigned a specific color (`vertex_colors`). These methods are
demonstrated in :ref:`sphx_glr_gallery_surface_texture_and_colors.py`.

In the case of "photorealistic" models, the color of each vertex is instead
determined by mapping a vertex to a point in an image called a texture using 2D
texture coordinates in the range [0, 1]. The color of each individual pixel is
smoothly interpolated (sampled) on the fly from the texture (the GPU makes this
interpolation very fast).

Napari does not (yet) support models with multiple textures or materials. If
the textures don't overlap, you can display them on separate meshes as shown in
this demo. If the textures do overlap, you may instead be able to combine the
textures as images. This relies on textures having the same texture
coordinates, and may require resizing the textures to match each other.

.. tags:: visualization-nD
"""
import os

import matplotlib.pyplot as plt
import pooch
from vispy.io import imread, read_mesh

import napari

###############################################################################
# Download the model
# ------------------
download = pooch.DOIDownloader(progressbar=True)
doi = '10.5281/zenodo.13380203'
tmp_dir = pooch.os_cache('napari-surface-texture-example')
os.makedirs(tmp_dir, exist_ok=True)
data_files = {
    'mesh': 'PocilloporaDamicornisSkin.obj',
    # "materials": "PocilloporaVerrugosaSkinCrop.mtl",  # not yet supported
    'Texture_0': 'PocilloporaDamicornisSkin_Texture_0.jpg',
    'GeneratedMat2': 'PocilloporaDamicornisSkin_GeneratedMat2.png',
}
data_to_path = {}
print(f'downloading data into {tmp_dir}')
for id_, file_name in data_files.items():
    res = pooch.retrieve(f'doi:{doi}/{file_name}', known_hash=None,  progressbar=True)
    data_to_path[id_] = res


###############################################################################
# Load the model
# --------------
# Next, read the model data from the .obj file. Currently napari/vispy do not
# support reading material properties (.mtl files) nor separate texture and
# vertex indices (i.e. repeated vertices). Normal vectors read from the file
# are also ignored and re-calculated from the faces.
vertices, faces, _normals, texcoords = read_mesh(data_to_path['mesh'])

###############################################################################
# Load the textures
# -----------------
# This model comes with two textures: `Texture_0` is generated from
# photogrammetry of the actual object, and `GeneratedMat2` is a generated
# material to fill in  parts of the model lacking photographic texture.
photo_texture = imread(data_to_path['Texture_0'])
generated_texture = imread(data_to_path['GeneratedMat2'])

###############################################################################
# This is what the texture images look like in 2D:
fig, axs = plt.subplots(1, 2)
axs[0].set_title(f'Texture_0 {photo_texture.shape}')
axs[0].imshow(photo_texture)
axs[0].set_xticks((0, photo_texture.shape[1]), labels=(0.0, 1.0))
axs[0].set_yticks((0, photo_texture.shape[0]), labels=(0.0, 1.0))
axs[1].set_title(f'GeneratedMat2 {generated_texture.shape}')
axs[1].imshow(generated_texture)
axs[1].set_xticks((0, generated_texture.shape[1]), labels=(0.0, 1.0))
axs[1].set_yticks((0, generated_texture.shape[0]), labels=(0.0, 1.0))
fig.show()

###############################################################################
# Create the napari layers
# ------------------------
# Next create two separate layers with the same mesh - once with each texture.
# In this example the texture coordinates happen to be the same for each
# texture, but this is not a strict requirement.
photo_texture_layer = napari.layers.Surface(
    (vertices, faces),
    texture=photo_texture,
    texcoords=texcoords,
    name='Texture_0',
)
generated_texture_layer = napari.layers.Surface(
    (vertices, faces),
    texture=generated_texture,
    texcoords=texcoords,
    name='GeneratedMat2',
)

###############################################################################
# Add the layers to a viewer
# --------------------------
# Finally, create the viewer and add the Surface layers.
# sphinx_gallery_thumbnail_number = 2
viewer = napari.Viewer(ndisplay=3)

viewer.add_layer(photo_texture_layer)
viewer.add_layer(generated_texture_layer)

viewer.camera.angles = (90.0, 0.0, -75.0)
viewer.camera.zoom = 75

if __name__ == '__main__':
    napari.run()



"""
nD multiscale image non-uniform
===============================

Displays an nD multiscale image

.. tags:: visualization-advanced
"""

import numpy as np
from skimage import data
from skimage.transform import pyramid_gaussian

import napari

# create multiscale from astronaut image
astronaut = data.astronaut()
base = np.tile(astronaut, (3, 3, 1))
multiscale = list(
    pyramid_gaussian(base, downscale=2, max_layer=3, channel_axis=-1)
)
multiscale = [
    np.array([p * (abs(3 - i) + 1) / 4 for i in range(6)]) for p in multiscale
]
print('multiscale level shapes: ', [p.shape for p in multiscale])

# add image multiscale
viewer = napari.Viewer()
layer = viewer.add_image(multiscale, multiscale=True)

if __name__ == '__main__':
    napari.run()



from __future__ import annotations

from typing import TYPE_CHECKING, Literal, cast

import numpy as np
from scipy.spatial.transform import Rotation as R

from napari._pydantic_compat import validator
from napari.utils.camera_orientations import (
    DEFAULT_ORIENTATION_TYPED,
    DepthAxisOrientation,
    Handedness,
    HorizontalAxisOrientation,
    HorizontalAxisOrientationStr,
    VerticalAxisOrientation,
    VerticalAxisOrientationStr,
)
from napari.utils.events import EventedModel
from napari.utils.misc import ensure_n_tuple

if TYPE_CHECKING:
    import numpy.typing as npt


class Camera(EventedModel):
    """Camera object modeling position and view of the camera.

    Attributes
    ----------
    center : 3-tuple
        Center of rotation for the camera.
        In 2D viewing the last two values are used.
    zoom : float
        Scale from canvas pixels to world pixels.
    angles : 3-tuple
        Euler angles of camera in 3D viewing (rx, ry, rz), in degrees.
        Only used during 3D viewing.
        Note that Euler angles's intrinsic degeneracy means different
        sets of Euler angles may lead to the same view.
    perspective : float
        Perspective (aka "field of view" in vispy) of the camera (if 3D).
    mouse_pan : bool
        If the camera interactive panning with the mouse is enabled or not.
    mouse_zoom : bool
        If the camera interactive zooming with the mouse is enabled or not.
    """

    # fields
    center: tuple[float, float, float] | tuple[float, float] = (
        0.0,
        0.0,
        0.0,
    )
    zoom: float = 1.0
    angles: tuple[float, float, float] = (0.0, 0.0, 0.0)
    perspective: float = 0
    mouse_pan: bool = True
    mouse_zoom: bool = True
    orientation: tuple[
        DepthAxisOrientation,
        VerticalAxisOrientation,
        HorizontalAxisOrientation,
    ] = DEFAULT_ORIENTATION_TYPED

    # validators
    @validator('center', 'angles', pre=True, allow_reuse=True)
    def _ensure_3_tuple(cls, v):
        return ensure_n_tuple(v, n=3)

    @property
    def view_direction(self) -> tuple[float, float, float]:
        """3D view direction vector of the camera.

        View direction is calculated from the Euler angles and returned as a
        3-tuple. This direction is in 3D scene coordinates, the world coordinate
        system for three currently displayed dimensions.
        """
        # once we're in scene-land, we pretend to be in xyz space (axes names don't
        # mean anything after all...) which simplifies the logic a lot.
        rotation = R.from_euler('xyz', self.angles, degrees=True)
        # view direction is given by the z component, but flipping the sign.
        # This is because the default view direction at angles (0, 0, 0) is (-1, 0, 0)
        return tuple(-rotation.as_matrix()[0])

    @property
    def up_direction(self) -> tuple[float, float, float]:
        """3D direction vector pointing up on the canvas.

        Up direction is calculated from the Euler angles and returned as a
        3-tuple. This direction is in 3D scene coordinates, the world coordinate
        system for three currently displayed dimensions.
        """
        # once we're in scene-land, we pretend to be in xyz space (axes names don't
        # mean anything after all...) which simplifies the logic a lot.
        rotation = R.from_euler('xyz', self.angles, degrees=True)
        # up direction is given by the y component, but flipping the sign.
        # This is because the default up direction at angles (0, 0, 0) is (0, -1, 0)
        return tuple(-rotation.as_matrix()[1])

    def set_view_direction(
        self,
        view_direction: tuple[float, float, float],
        up_direction: tuple[float, float, float] = (0, -1, 0),
    ):
        """Set camera angles from direction vectors.

        Both the view direction and the up direction are specified in 3D scene
        coordinates, the world coordinate system for three currently displayed
        dimensions.

        The provided up direction must not be parallel to the provided
        view direction. The provided up direction does not need to be orthogonal
        to the view direction. The final up direction will be a vector orthogonal
        to the view direction, aligned with the provided up direction.

        Parameters
        ----------
        view_direction : 3-tuple of float
            The desired view direction vector in 3D scene coordinates, the world
            coordinate system for three currently displayed dimensions.
        up_direction : 3-tuple of float
            A direction vector which will point upwards on the canvas. Defaults
            to (0, -1, 0) unless the view direction is parallel to the y-axis,
            in which case will default to (-1, 0, 0).
        """
        # project up onto view so we can remove the parallel component
        projection = np.dot(up_direction, view_direction) * np.array(
            view_direction
        )
        up_direction_arr = np.asarray(up_direction) - projection

        view_direction_arr = np.asarray(view_direction) / np.linalg.norm(
            view_direction
        )
        up_direction_arr = up_direction_arr / np.linalg.norm(up_direction_arr)
        right_direction = np.cross(up_direction_arr, view_direction_arr)

        # once we're in scene-land, we pretend to be in xyz space (axes names don't
        # mean anything after all...) which simplifies the logic a lot. We also
        # flip all signs (see explanations in self.view_direction, and self.up_direction)
        matrix = -np.array(
            (view_direction_arr, up_direction_arr, right_direction)
        )
        self.angles = R.from_matrix(matrix).as_euler('xyz', degrees=True)

    def calculate_nd_view_direction(
        self, ndim: int, dims_displayed: tuple[int, ...]
    ) -> npt.NDArray[np.float64] | None:
        """Calculate the nD view direction vector of the camera.

        Parameters
        ----------
        ndim : int
            Number of dimensions in which to embed the 3D view vector.
        dims_displayed : Tuple[int]
            Dimensions in which to embed the 3D view vector.

        Returns
        -------
        view_direction_nd : np.ndarray
            nD view direction vector as an (ndim, ) ndarray
        """
        if len(dims_displayed) != 3:
            return None
        view_direction_nd = np.zeros(ndim)
        view_direction_nd[list(dims_displayed)] = self.view_direction
        return view_direction_nd

    def calculate_nd_up_direction(
        self, ndim: int, dims_displayed: tuple[int, ...]
    ) -> np.ndarray | None:
        """Calculate the nD up direction vector of the camera.

        Parameters
        ----------
        ndim : int
            Number of dimensions in which to embed the 3D view vector.
        dims_displayed : Tuple[int]
            Dimensions in which to embed the 3D view vector.

        Returns
        -------
        up_direction_nd : np.ndarray
            nD view direction vector as an (ndim, ) ndarray
        """
        if len(dims_displayed) != 3:
            return None
        up_direction_nd = np.zeros(ndim)
        up_direction_nd[list(dims_displayed)] = self.up_direction
        return up_direction_nd

    @property
    def orientation2d(
        self,
    ) -> tuple[VerticalAxisOrientation, HorizontalAxisOrientation]:
        return self.orientation[1:]

    @orientation2d.setter
    def orientation2d(
        self,
        value: tuple[
            VerticalAxisOrientation | VerticalAxisOrientationStr,
            HorizontalAxisOrientation | HorizontalAxisOrientationStr,
        ],
    ) -> None:
        self.orientation = (
            self.orientation[0],
            VerticalAxisOrientation(value[0]),
            HorizontalAxisOrientation(value[1]),
        )

    @property
    def handedness(self) -> Handedness:
        """Right or left-handedness of the current orientation."""
        # we know default orientation is right-handed, so an odd number of
        # differences from default means left-handed.
        diffs = [
            self.orientation[i] != DEFAULT_ORIENTATION_TYPED[i]
            for i in range(3)
        ]
        if sum(diffs) % 2 != 0:
            return Handedness.LEFT
        return Handedness.RIGHT

    def _vispy_flipped_axes(
        self, ndisplay: Literal[2, 3] = 2
    ) -> tuple[int, int, int]:
        # Note: the Vispy axis order is xyz, or horizontal, vertical, depth,
        # while the napari axis order is zyx / plane-row-column, or depth, vertical,
        # horizontal — i.e. it is exactly inverted. This switch happens when data
        # is passed from napari to Vispy, usually with a transposition. In the camera
        # models, this means that the order of these orientations appear in the
        # opposite order to that in napari.components.Camera.
        #
        # Note that the default Vispy camera orientations come from Vispy, not from us.
        vispy_default_orientation = (
            ('right', 'up', 'towards')
            if ndisplay == 2
            else ('right', 'down', 'away')
        )

        # Vispy uses xyz coordinates; napari uses zyx coordinates. We therefore
        # start by inverting the order of coordinates coming from the napari
        # camera model:
        orientation_xyz = self.orientation[::-1]
        # The Vispy camera flip is a tuple of three ints in {0, 1}, indicating
        # whether they are flipped relative to the Vispy default.
        return cast(
            tuple[int, int, int],
            tuple(
                int(ori != default_ori)
                for ori, default_ori in zip(
                    orientation_xyz, vispy_default_orientation, strict=True
                )
            ),
        )



"""
Check state of vendored modules.
"""

import shutil
import sys
from pathlib import Path
from subprocess import check_output
from typing import List


TOOLS_PATH = Path(__file__).parent
REPO_ROOT_PATH = TOOLS_PATH.parent
VENDOR_FOLDER = "_vendor"
NAPARI_FOLDER = "napari"


def _clone(org, reponame, tag):
    repo_path = TOOLS_PATH / reponame
    if repo_path.is_dir():
        shutil.rmtree(repo_path)

    check_output(
        [
            "git",
            "clone",
            '--depth',
            '1',
            '--branch',
            tag,
            f"https://github.com/{org}/{reponame}",
        ],
        cwd=TOOLS_PATH,
    )

    return repo_path


def check_vendored_files(
    org: str, reponame: str, tag: str, source_paths: List[Path], target_path: Path
) -> str:
    repo_path = _clone(org, reponame, tag)
    vendor_path = REPO_ROOT_PATH / 'src' / NAPARI_FOLDER / target_path
    for s in source_paths:
        shutil.copy(repo_path / s, vendor_path)
    return check_output(["git", "diff"], cwd=vendor_path).decode("utf-8")


def check_vendored_module(org: str, reponame: str, tag: str) -> str:
    """
    Check if the vendored module is up to date.

    Parameters
    ----------
    org : str
        The github organization name.
    reponame : str
        The github repository name.
    tag : str
        The github tag.

    Returns
    -------
    str
        Returns the diff if the module is not up to date or an empty string
        if it is.
    """
    repo_path = _clone(org, reponame, tag)

    vendor_path = REPO_ROOT_PATH / 'src' / NAPARI_FOLDER / VENDOR_FOLDER / reponame
    if vendor_path.is_dir():
        shutil.rmtree(vendor_path)

    shutil.copytree(repo_path / reponame, vendor_path)
    shutil.copy(repo_path / "LICENSE", vendor_path)
    shutil.rmtree(repo_path, ignore_errors=True)

    return check_output(["git", "diff"], cwd=vendor_path).decode("utf-8")


def main():
    CI = '--ci' in sys.argv
    print("\n\nChecking vendored modules\n")
    vendored_modules = []
    for org, reponame, tag, source, target in [
        ("albertosottile", "darkdetect", "master", None, None),
        (
            "matplotlib",
            "matplotlib",
            "main",
            [
                # this file seem to be post 3.0.3 but pre 3.1
                # plus there may have been custom changes.
                # 'lib/matplotlib/colors.py',
                #
                # this file seem much more recent, but is touched much more rarely.
                # it is at least from 3.2.1 as the turbo colormap is present and
                # was added in matplotlib in 3.2.1
                'lib/matplotlib/_cm_listed.py'
            ],
            'utils/colormaps/vendored/',
        ),
    ]:
        print(f"\n * Checking '{org}/{reponame}'\n")
        if source is None:
            diff = check_vendored_module(org, reponame, tag)
        else:
            diff = check_vendored_files(
                org, reponame, tag, [Path(s) for s in source], Path(target)
            )

        if diff:
            vendored_modules.append((org, reponame, diff))

    if CI:
        with open(TOOLS_PATH / "vendored_modules.txt", "w") as f:
            f.write(" ".join(f"{org}/{reponame}" for org, reponame, _ in vendored_modules))
        sys.exit(0)
    if vendored_modules:
        print("\n\nThe following vendored modules are not up to date:\n")
        for org, reponame, _diff in vendored_modules:
            print(f"\n * {org}/{reponame}\n")
        sys,exit(1)


if __name__ == "__main__":
    main()



"""
Add vectors color by angle
==========================

This example generates a set of vectors in a spiral pattern.
The color of the vectors is mapped to their 'angle' feature.

.. tags:: visualization-advanced
"""

import numpy as np
from skimage import data

import napari

# create the viewer and window
viewer = napari.Viewer()

layer = viewer.add_image(data.camera(), name='photographer')

# sample vector coord-like data
n = 300
pos = np.zeros((n, 2, 2), dtype=np.float32)
phi_space = np.linspace(0, 4 * np.pi, n)
radius_space = np.linspace(0, 100, n)

# assign x-y position
pos[:, 0, 0] = radius_space * np.cos(phi_space) + 300
pos[:, 0, 1] = radius_space * np.sin(phi_space) + 256

# assign x-y projection
pos[:, 1, 0] = 2 * radius_space * np.cos(phi_space)
pos[:, 1, 1] = 2 * radius_space * np.sin(phi_space)

# make the angle feature, range 0-2pi
angle = np.mod(phi_space, 2 * np.pi)

# create a feature that is true for all angles  > pi
pos_angle = angle > np.pi

# create the features dictionary.
features = {
    'angle': angle,
    'pos_angle': pos_angle,
}

# add the vectors
layer = viewer.add_vectors(
    pos,
    edge_width=3,
    features=features,
    edge_color='angle',
    edge_colormap='husl',
    name='vectors'
)

# set the edge color mode to colormap
layer.edge_color_mode = 'colormap'

if __name__ == '__main__':
    napari.run()



"""
Labels 3D
=========

View 3D labels.

.. tags:: visualization-nD
"""


import numpy as np
from scipy import ndimage as ndi
from skimage import data, filters, morphology

import napari

cells3d = data.cells3d()
viewer = napari.Viewer()
membranes_layer, nuclei_layer = viewer.add_image(
    cells3d, channel_axis=1, name=['membranes', 'nuclei']
)
membrane, nuclei = cells3d.transpose((1, 0, 2, 3)) / np.max(cells3d)
edges = filters.scharr(nuclei)
denoised = ndi.median_filter(nuclei, size=3)
thresholded = denoised > filters.threshold_li(denoised)
cleaned = morphology.remove_small_objects(
    morphology.remove_small_holes(thresholded, 20**3),
    20**3,
)

segmented = ndi.label(cleaned)[0]
# maxima = ndi.label(morphology.local_maxima(filters.gaussian(nuclei, sigma=10)))[0]
# markers_big = morphology.dilation(maxima, morphology.ball(5))

# segmented = segmentation.watershed(
#     edges,
#     markers_big,
#     mask=cleaned,
# )

labels_layer = viewer.add_labels(segmented)

viewer.dims.ndisplay = 3

if __name__ == '__main__':
    napari.run()



"""
3D image plane rendering
========================

Display one 3D image layer and display it as a plane
with a simple widget for modifying plane parameters.

.. tags:: visualization-advanced, gui, layers
"""
import numpy as np
from skimage import data

import napari
from napari.utils.translations import trans

viewer = napari.Viewer(ndisplay=3)

# add a 3D image
blobs = data.binary_blobs(
    length=64, volume_fraction=0.1, n_dim=3
).astype(np.float32)
image_layer = viewer.add_image(
    blobs, rendering='mip', name='volume', blending='additive', opacity=0.25
)

# add the same 3D image and render as plane
# plane should be in 'additive' blending mode or depth looks all wrong
plane_parameters = {
    'position': (32, 32, 32),
    'normal': (0, 1, 0),
    'thickness': 10,
}

plane_layer = viewer.add_image(
    blobs,
    rendering='average',
    name='plane',
    depiction='plane',
    blending='additive',
    opacity=0.5,
    plane=plane_parameters
)
viewer.axes.visible = True
viewer.camera.angles = (45, 45, 45)
viewer.camera.zoom = 5
viewer.text_overlay.text = trans._(
    """
shift + click and drag to move the plane
press 'x', 'y' or 'z' to orient the plane along that axis around the cursor
press 'o' to orient the plane normal along the camera view direction
press and hold 'o' then click and drag to make the plane normal follow the camera
"""
)
viewer.text_overlay.visible = True
if __name__ == '__main__':
    napari.run()



"""
magicgui viewer
===============

Example showing how to access the current viewer from a function widget.

.. tags:: gui
"""

import napari


# annotating a parameter as `napari.Viewer` will automatically provide
# the viewer that the function is embedded in, when the function is added to
# the viewer with add_function_widget.
def my_function(viewer: napari.Viewer):
    print(viewer, f'with {len(viewer.layers)} layers')


viewer = napari.Viewer()
# Add our magic function to napari
viewer.window.add_function_widget(my_function)

if __name__ == '__main__':
    napari.run()



"""
Interactive scripting
=====================

.. tags:: interactivity
"""

import time

import numpy as np

import napari
from napari.qt import thread_worker

# create the viewer with an image
data = np.random.random((512, 512))
viewer = napari.Viewer()
layer = viewer.add_image(data)

def update_layer(data):
    layer.data = data

@thread_worker(connect={'yielded': update_layer})
def create_data(*, update_period, num_updates):
    # number of times to update
    for _k in range(num_updates):
        yield np.random.random((512, 512))
        time.sleep(update_period)

create_data(update_period=0.05, num_updates=50)

if __name__ == '__main__':
    napari.run()



import numpy.typing as npt

from napari.layers._scalar_field._slice import _ScalarFieldSliceRequest
from napari.layers.base._base_constants import BaseProjectionMode
from napari.types import ArrayLike


class _LabelsSliceRequest(_ScalarFieldSliceRequest):
    @staticmethod
    def _project_slice(
        data: ArrayLike, axis: tuple[int, ...], mode: BaseProjectionMode
    ) -> npt.NDArray:
        """Project a thick slice along axis based on mode."""
        raise NotImplementedError



"""
nD vectors
==========

Display two vectors layers ontop of a 4-D image layer. One of the vectors
layers is 3D and "sliced" with a different set of vectors appearing on
different 3D slices. Another is 2D and "broadcast" with the same vectors
apprearing on each slice.

.. tags:: visualization-nD
"""

import numpy as np
from skimage import data

import napari

blobs = np.stack(
    [
        data.binary_blobs(
            length=128, blob_size_fraction=0.05, n_dim=3, volume_fraction=f
        )
        for f in np.linspace(0.05, 0.5, 10)
    ],
    axis=0,
)
viewer = napari.Viewer()
layer = viewer.add_image(blobs.astype(float))

# sample vector coord-like data
n = 200
pos = np.zeros((n, 2, 2), dtype=np.float32)
phi_space = np.linspace(0, 4 * np.pi, n)
radius_space = np.linspace(0, 20, n)

# assign x-y position
pos[:, 0, 0] = radius_space * np.cos(phi_space) + 64
pos[:, 0, 1] = radius_space * np.sin(phi_space) + 64

# assign x-y projection
pos[:, 1, 0] = 2 * radius_space * np.cos(phi_space)
pos[:, 1, 1] = 2 * radius_space * np.sin(phi_space)

planes = np.round(np.linspace(0, 128, n)).astype(int)
planes = np.concatenate(
    (planes.reshape((n, 1, 1)), np.zeros((n, 1, 1))), axis=1
)
vectors = np.concatenate((planes, pos), axis=2)

# add the sliced vectors
layer = viewer.add_vectors(
    vectors, edge_width=0.4, name='sliced vectors', edge_color='blue'
)

viewer.dims.ndisplay = 3

if __name__ == '__main__':
    napari.run()



"""
Progress bar minimal
====================

This file provides minimal working examples of progress bars in
the napari viewer.

.. tags:: gui
"""

from random import choice
from time import sleep

import numpy as np
from qtpy.QtWidgets import QPushButton, QVBoxLayout, QWidget

import napari
from napari.utils import cancelable_progress, progress


def process(im_slice):
    # do something with your image slice
    sleep(0.4)

def iterable():
    """using progress as a wrapper for iterables
    """
    my_stacked_volume = np.random.random((5, 4, 500, 500))
    # we can wrap any iterable object in `progress` and see a progress
    # bar in the viewer
    for im_slice in progress(my_stacked_volume):
        process(im_slice)

def iterable_w_context():
    """using progress with a context manager
    """
    my_stacked_volume = np.random.random((5, 4, 500, 500))
    # progress provides a context manager we can use for automatic
    # teardown of our widget once iteration is complete. Wherever
    # possible, we should *always* use progress within a context
    with progress(my_stacked_volume) as pbr:
        for i, im_slice in enumerate(pbr):
            # using a context manager also allows us to manipulate
            # the progress object e.g. by setting a description
            pbr.set_description(f'Slice {i}')

            # we can group progress bars together in the viewer
            # by passing a parent progress bar to new progress
            # objects' nest_under attribute
            for channel in progress(im_slice, nest_under=pbr):
                process(channel)

def indeterminate():
    """By passing a total of 0, we can have an indeterminate progress bar
    """

    # note progress(total=0) is equivalent to progress()
    with progress(total=0) as pbr:
        x = 0
        while x != 42:
            pbr.set_description(f'Processing {x}')
            x = choice(range(100))
            sleep(0.05)

def arbitrary_steps():
    """We can manually control updating the value of the progress bar.
    """
    with progress(total=4) as pbr:
        sleep(3)
        pbr.set_description('Step 1 Complete')
        # manually updating the progress bar by 1
        pbr.update(1)

        sleep(1)
        pbr.set_description('Step 2 Complete')
        pbr.update(1)

        sleep(2)
        pbr.set_description('Processing Complete!')
        # we can manually update by any number of steps
        pbr.update(2)

        # sleeping so we can see full completion
        sleep(1)


def cancelable_iterable():
    """We can allow expensive computations to be cancelable
    """
    # Note that if canceled, for loop will terminate prematurely
    # You can use cancel_callback to close files, clean up state, etc
    # if the user cancels the operation.
    def cancel_callback():
        print('Operation canceled - cleaning up!')
    for _ in cancelable_progress(range(100), cancel_callback=cancel_callback):
        np.random.rand(128, 128, 128).mean(0)

viewer = napari.Viewer()
button_layout = QVBoxLayout()

iterable_btn = QPushButton('Iterable')
iterable_btn.clicked.connect(iterable)
button_layout.addWidget(iterable_btn)

iterable_context_btn = QPushButton('Iterable With Context')
iterable_context_btn.clicked.connect(iterable_w_context)
button_layout.addWidget(iterable_context_btn)

indeterminate_btn = QPushButton('Indeterminate')
indeterminate_btn.clicked.connect(indeterminate)
button_layout.addWidget(indeterminate_btn)

steps_btn = QPushButton('Arbitrary Steps')
steps_btn.clicked.connect(arbitrary_steps)
button_layout.addWidget(steps_btn)

cancel_iter_btn = QPushButton('Cancelable Iterable')
cancel_iter_btn.clicked.connect(cancelable_iterable)
button_layout.addWidget(cancel_iter_btn)

pbar_widget = QWidget()
pbar_widget.setLayout(button_layout)
pbar_widget.setObjectName('Progress Examples')

viewer.window.add_dock_widget(pbar_widget)
# showing the activity dock so we can see the progress bars
viewer.window._status_bar._toggle_activity_dock(True)

if __name__ == '__main__':
    napari.run()



"""
nD vectors image
================

This example generates an image of vectors
Vector data is an array of shape (M, N, P, 3)
Each vector position is defined by an (x-proj, y-proj, z-proj) element
which are vector projections centered on a pixel of the MxNxP grid

.. tags:: visualization-nD
"""

import numpy as np

import napari

# create the viewer and window
viewer = napari.Viewer()

m = 10
n = 20
p = 40

image = 0.2 * np.random.random((m, n, p)) + 0.5
layer = viewer.add_image(image, contrast_limits=[0, 1], name='background')

# sample vector image-like data
# n x m grid of slanted lines
# random data on the open interval (-1, 1)
pos = np.random.uniform(-1, 1, size=(m, n, p, 3))
print(image.shape, pos.shape)

# add the vectors
vect = viewer.add_vectors(pos, edge_width=0.2, length=2.5)

if __name__ == '__main__':
    napari.run()



"""
Paint nD
========

Display a 4D labels layer and paint only in 3D.

This is useful e.g. when proofreading segmentations within a time series.

.. tags:: analysis
"""

import numpy as np
from skimage import data

import napari

blobs = np.stack(
    [
        data.binary_blobs(
            length=128, blob_size_fraction=0.05, n_dim=3, volume_fraction=f
        )
        for f in np.linspace(0.05, 0.5, 10)
    ],
    axis=0,
)
viewer = napari.Viewer()
layer = viewer.add_image(blobs.astype(float), rendering='attenuated_mip')
labels = viewer.add_labels(np.zeros_like(blobs, dtype=np.int32))
labels.n_edit_dimensions = 3
labels.brush_size = 15
labels.mode = 'paint'

if __name__ == '__main__':
    napari.run()



"""
Glasbey colormap
================

A long-requested feature [1]_ for napari was to display labels/segmentation
layers using a well-known discrete colormap such as Glasbey [2]_.

In this example, we demonstrate displaying segmentations using custom colormaps
with the help of the glasbey Python library [3]_, which you can install with
your favorite Python package manager, such as pip or conda. We display a
segmentation using the napari built-in labels colormap, the original Glasbey
colormap, a more modern version produced by limiting the lightness and chroma
and optimizing for colorblind-safety, and finally with the matplotlib tab10
colormap.

.. [1] https://github.com/napari/napari/issues/454
.. [2] Colour displays for categorical images. Chris Glasbey, Gerie van der
       Heijden, Vivian F. K. Toh, and Alision Gray. (2007)
       DOI:10.1002/col.20327
.. [3] https://github.com/lmcinnes/glasbey

.. tags:: layers, visualization-basic
"""
import glasbey
import numpy as np
from skimage import data
from skimage.filters import threshold_otsu
from skimage.measure import label
from skimage.morphology import closing, remove_small_objects
from skimage.segmentation import clear_border

import napari

image = data.coins()[50:-50, 50:-50]

###############################################################################
# First, we segment the image.

# apply threshold
thresh = threshold_otsu(image)
bw = closing(image > thresh, footprint=np.ones((4, 4), dtype=bool))

# remove artifacts connected to image border
cleared = remove_small_objects(clear_border(bw), 20)

# label image regions
label_image = label(cleared).astype("uint8")

###############################################################################
# Then, we create two color palettes using the glasbey library. One with the
# original glasbey parameters and 256 colors, and a more modern one with better
# lightness and chroma bounds for a less glary look.

# original glasbey
glas = glasbey.create_palette(256)

# more optimized glasbey
glas19mid = glasbey.create_palette(
        19,
        lightness_bounds=(20, 60), chroma_bounds=(40, 50),
        colorblind_safe=True,
        )

###############################################################################
# Finally, we display the coins image and the overlaid segmentation. We do this
# in two viewers to show both colormaps.

viewer, image_layer = napari.imshow(image, name='coins')

# add the labels
label_layer_glas = viewer.add_labels(
        label_image, name='segmentation', colormap=glas
        )

viewer2, image_layer2 = napari.imshow(image, name='coins')

label_layer_modern = viewer2.add_labels(
        label_image, name='segmentation-glasbey-19-mid-chroma', colormap=glas19mid,
        )


if __name__ == '__main__':
    napari.run()



import json
import pathlib

data = json.loads(
    (pathlib.Path(__file__).parent / 'string_list.json').read_text()
)


SKIP_FOLDERS = data['SKIP_FOLDERS']
SKIP_FILES = data['SKIP_FILES']
SKIP_WORDS_GLOBAL = data['SKIP_WORDS_GLOBAL']
SKIP_WORDS = data['SKIP_WORDS']



import cProfile
import io
import pstats
from pstats import SortKey

import numpy as np

from napari.layers import Points

"""
This script was useful for testing how the performance of
Points._set_view_slice changed with different implementations during async
development.
"""

np.random.seed(0)

n = 65536
data = np.random.random((n, 2))
s = io.StringIO()

reps = 100

# Profiling
with cProfile.Profile() as pr:
    for _ in range(reps):
        layer = Points(data)
        layer._set_view_slice()

sortby = SortKey.CUMULATIVE
ps = pstats.Stats(pr, stream=s).sort_stats(sortby)
ps.print_stats(0.05)
print(s.getvalue())

# pr.dump_stats("result.pstat")
